#!/usr/bin/env
"""
Field line tracer specifics
author: daniel.boeckenhoff@ipp.mpg.de
"""

import w7x
from w7x.components import MeshedModel
import rna
import tfields

import numpy as np
import logging
import os
import functools


class Points3D(w7x.core.BasePoints3D):
    ws_server = w7x.Server.addr_field_line_server


class Base(w7x.core.Base):
    """
    Field Line Server Base
    """

    ws_server = w7x.Server.addr_field_line_server


class ConnectionLength(Points3D):
    """Class to Convert ConnectionLengthResult in.
    Make it easy to handle info about the different parts you gave it.
    The Hit points are the content of the np.ndarray (Points3D)

    Note:

    Args:
        connectionLengthResult

    Attributes:
        x1 (list): coordinate x
        x2 (list): coordinate y
        x3 (list): coordinate z
        lengths (list)
        parts (list): part ids refering to mm_ids index
        elements (list): element ids refering to face index + 1

    Examples:
        Adding up two instances
        >>> from w7x.flt import ConnectionLength
        >>> import numpy as np
        >>> c = ConnectionLength.create_test()
        >>> c.parts
        [0, 1, 2, 3, 4]
        >>> d = ConnectionLength.create_test()
        >>> b = ConnectionLength.merged(c, d)
        >>> b.parts
        [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]
        >>> b
        ConnectionLength([[  1.        ,   0.        ,   0.97438334],
                          [  0.        ,   1.        ,   0.9579201 ],
                          [-80.        ,   1.        ,   1.9573    ],
                          [-80.        ,  -1.        ,   0.22222222],
                          [  0.        ,  -1.        ,   0.9579201 ],
                          [  1.        ,   0.        ,   0.97438334],
                          [  0.        ,   1.        ,   0.9579201 ],
                          [-80.        ,   1.        ,   1.9573    ],
                          [-80.        ,  -1.        ,   0.22222222],
                          [  0.        ,  -1.        ,   0.9579201 ]])
        >>> assert len(b.lengths) == 10
        >>> b.lengths[:4]
        [899.895318339, 2826.55635289, 2.55635289, 222.822222339]
        >>> b.elements
        [44301, 20291, 2, 22222, 20291, 44301, 20291, 2, 22222, 20291]
        >>> len(b)
        10

        Retrieving Groups for all the components
        >>> b.mm_ids = [15, 16, 17, 18, 19]
        >>> assert b.hits(15).equal([[ 1.,  0.,  0.97438334],
        ...                          [ 1.,  0.,  0.97438334]],
        ...                         atol=1e-6, rtol=1e-6)
        >>> p = b.hits(18)
        >>> p.to_segment_one()
        >>> assert p.equal([[ 65.3091448 , -46.21380319, 0.22222222],
        ...                 [ 65.3091448 , -46.21380319, 0.22222222]],
        ...                atol=1e-6, rtol=1e-6)

        Saving and loading like in Points3D
        >>> from tempfile import NamedTemporaryFile
        >>> outFile = NamedTemporaryFile(suffix='.npz')
        >>> c.save(outFile.name)
        >>> _ = outFile.seek(0)
        >>> c1 = ConnectionLength.load(outFile.name)
        >>> assert np.all(c == c1)
        >>> assert np.array_equal(c.parts, [0, 1, 2, 3, 4])
        >>> assert np.array_equal(c.parts, c1.parts)

    """

    __slots__ = ["coord_sys", "name", "parts", "elements", "lengths", "mm_ids"]
    __slot_setters__ = [
        tfields.bases.get_coord_system_name,
        None,
        None,
        None,
        None,
        "_set_mm_ids",
    ]

    def __new__(cls, tensors, **kwargs):
        # TODO(@dboe): this should be unneccesarry if u just use a field, idiot
        if isinstance(tensors, list):
            if w7x.is_w7x_instance(tensors, cls.get_ws_class()) or all(
                [
                    hasattr(con, attr)
                    for con in tensors
                    for attr in ["x", "y", "z", "length", "part", "element"]
                ]
            ):
                points = []
                parts = []
                elements = []
                lengths = []
                for con in tensors:
                    points.append([con.x, con.y, con.z])
                    lengths.append(con.length)
                    parts.append(con.part)
                    elements.append(con.element)
                tensors = points
                kwargs["parts"] = parts
                kwargs["elements"] = elements
                kwargs["lengths"] = lengths

        parts = kwargs.pop("parts", None)
        elements = kwargs.pop("elements", None)
        lengths = kwargs.pop("lengths", None)
        _mm_ids = kwargs.pop("_mm_ids", None)

        if issubclass(type(tensors), ConnectionLength):
            parts = parts or tensors.parts
            elements = elements or tensors.elements
            lengths = lengths or tensors.lengths
            _mm_ids = _mm_ids or tensors._mm_ids

        obj = super(ConnectionLength, cls).__new__(cls, tensors, **kwargs)
        obj.parts = parts
        obj.elements = elements
        obj.lengths = lengths
        obj._mm_ids = _mm_ids
        return obj

    @classmethod
    def merged(cls, *objects, **kwargs):
        """
        Overload merged for taking care of parts, elements and lenghts
            atrributes
        """
        if not all([isinstance(o, cls) for o in objects]):
            # TODO: could allow if all faceScalars are none
            raise TypeError(
                "Merge constructor only accepts {cls} instances.".format(**locals())
            )

        inst = super(ConnectionLength, cls).merged(*objects, **kwargs)

        kwargs["parts"] = kwargs.pop(
            "parts",
            functools.reduce(lambda x, y: x + y, [obj.parts for obj in objects]),
        )
        kwargs["elements"] = kwargs.pop(
            "elements",
            functools.reduce(lambda x, y: x + y, [obj.elements for obj in objects]),
        )
        kwargs["lengths"] = kwargs.pop(
            "lengths",
            functools.reduce(lambda x, y: x + y, [obj.lengths for obj in objects]),
        )
        inst = cls.__new__(cls, inst, **kwargs)
        return inst

    @classmethod
    def create_test(cls):
        """
        Create a test class filled with some data.
        """

        class C:
            def __init__(self, **kwargs):
                self.__dict__.update(kwargs)

        a = []
        a.append(
            C(length=899.895318339, x=1.0, y=0, z=0.974383343584, part=0, element=44301)
        )
        a.append(
            C(length=2826.55635289, x=0, y=1, z=0.95792010423, part=1, element=20291)
        )
        a.append(C(length=2.55635289, x=-80, y=1, z=1.9573, part=2, element=2))
        a.append(
            C(
                length=222.822222339,
                x=-80,
                y=-1,
                z=0.222222222224,
                part=3,
                element=22222,
            )
        )
        a.append(
            C(length=2826.55635289, x=0, y=-1, z=0.95792010423, part=4, element=20291)
        )
        obj = cls(a)
        obj.mm_ids = list(range(5))
        return obj

    def _set_mm_ids(self, mm_ids):
        """
        slot setter for mm_ids
        """
        if self.parts is not None and mm_ids is not None:
            if max([p if p is not None else 0 for p in self.parts]) >= len(mm_ids):
                raise ValueError("Too many parts for the mm_ids given.")
        return mm_ids

    def hits(self, *mm_ids):
        """
        return the hit points of the field lines on the components of the
        connection Results for the given mm_ids.
        """
        mm_ids = set(mm_ids)
        mask = []
        for part in self.parts:
            if part < 0:  # All invalid parts get False, rest True
                mask.append(False)
            elif self.mm_ids[part] not in mm_ids:
                mask.append(False)
            else:
                mask.append(True)
        p = Points3D(self[mask])
        return p

    def meshed_models(self, mm_ids=None):
        if mm_ids is None:
            mm_ids = self.mm_ids
        elif not set(mm_ids).issubset(self.mm_ids):
            raise ValueError("mm_ids must be subset of self.mm_ids.")

        meshed_models = []
        for mm_id in mm_ids:
            meshed_models.append(MeshedModel.from_mm_id(mm_id))
        return meshed_models


class ComponentLoad(object):
    """
    Collection class for flt.ComponentLoad instances
    Provides methods for choosing specific components only

    Args:
        *component_load (webservice ComponentLoad):
            e.g.:
                (ComponentLoad){
                    id = 0
                    events = 1
                    elements[] = [
                                  (ElementLoad){
                                      id = 103638
                                      events = 1
                                      area = 0.000175667334795
                                  }
                                  ]
                }
    Examples:
        >>> from w7x.flt import ComponentLoad
        >>> l = ComponentLoad.create_test()
        >>> l.mm_ids = [181, 182, 183, 184, 185]
        >>> l.load_dict()[181]['events'] == [1, 42, 1]
        True

    """

    def __init__(self, component_load):
        self.componentIdDict = {}
        for cl in component_load:
            if cl.id not in self.componentIdDict:
                self.componentIdDict[cl.id] = {}
            for element in cl.elements:
                if element.id not in self.componentIdDict[cl.id]:
                    self.componentIdDict[cl.id][element.id] = element.events
                else:
                    self.componentIdDict[cl.id][element.id] += element.events

    @property
    def mm_ids(self):
        """
        database IDs from ComponentsDB
        """
        return self._mm_ids

    @mm_ids.setter
    def mm_ids(self, mm_ids):
        self._mm_ids = mm_ids

    def _get_element_ids(self, component_id):
        return sorted(self.componentIdDict[component_id].keys())

    def _get_element_events(self, component_id):
        return [
            self.componentIdDict[component_id][eId]
            for eId in self._get_element_ids(component_id)
        ]

    def load_dict(self):
        """
        Returns:
            dictionary with following structure:
                {<mm_id>: {'elementIds': <list of ids of elements/triangles>,
                           'events': <list with the number of occuring events
                                corresponding to an element/triangle>},
                 <mm_id2>: ...}
        """
        mm_id_load_dict = {}
        if self.mm_ids is None:
            raise ValueError("Please set mm_ids before.")
        for component_id in sorted(self.componentIdDict.keys()):
            mm_id_load_dict[self.mm_ids[component_id]] = {
                "elementIds": self._get_element_ids(component_id),
                "events": self._get_element_events(component_id),
            }
        return mm_id_load_dict

    @classmethod
    def create_test(cls):
        """
        Create a test class filled with some data.
        """

        class C:
            def __init__(self, **kwargs):
                self.__dict__.update(kwargs)

        a = []
        a.append(
            C(
                id=0,
                events=3,
                elements=[
                    C(id=103638, events=1, area=0.00017),
                    C(id=123456, events=1, area=0.00017),
                    C(id=134679, events=1, area=0.00017),
                ],
            )
        )
        a.append(C(id=3, events=21, elements=[C(id=123, events=21, area=0.00017)]))
        a.append(C(id=0, events=41, elements=[C(id=123456, events=41, area=0.00017)]))
        return cls(a)


class Curve3D(Base):
    pass


def interprete_mm_id(mm_id):
    """
    Examples:
        >>> from w7x.flt import interprete_mm_id

        mm_id defined in web service
        >>> interprete_mm_id(30)
        (30, None)

        mm_id derived from web service
        >>> interprete_mm_id(420000030)
        (30, 42)

        Own independent mm_id
        >>> interprete_mm_id(421000000)
        (None, 42)

        Own independent mm_id
        >>> interprete_mm_id(1000)
        (None, 1000)

    Returns:
        tuple with 2 arguments:
            1) ws_mm_id (int or None): mm_id from web service
                if derived from web service definition
            2) own_mm_id (int or None): user defined id

    """
    ws_mm_id = None
    own_mm_id = None
    if mm_id < 1e6:
        if mm_id < 1e3:
            ws_mm_id = mm_id
        else:
            own_mm_id = mm_id
    else:
        headTmp = mm_id // int(1e6)
        head = headTmp // 10
        mid = headTmp - head * 10
        tail = mm_id - headTmp * int(1e6)

        if mid == 1:
            # not derived
            if tail != 0:
                raise ValueError(
                    "If mid index is 1, The mm_id"
                    "should not be derived by definition!"
                )
        elif mid == 0:
            ws_mm_id = tail
        else:
            raise ValueError("Mid digit must be either 1 or 0.")
        own_mm_id = head

    return ws_mm_id, own_mm_id


def component_module(mm_id):
    """
    Returns:
        int: which module the geometry is in
    """
    compDBServer = w7x.get_server(w7x.Server.addr_components_db)
    if mm_id >= 1000:
        mm_id = int(mm_id / 1000)
    componentInfo = compDBServer.service.getComponentInfo(mm_id)[0]
    return int(componentInfo.location.replace("m ", ""))


class LCFSSettings(Base):
    prop_defaults = {
        "LCFSLeftX": None,
        "LCFSRightX": None,
        "LCFSNumPoints": 40,
        "LCFSThreshold": 1000.0,
        "LCFSAccuracy": 0.001,
    }
    prop_order = ["LCFSLeftX", "LCFSRightX"]


class CylindricalGrid(Base):
    """
    Order of grid in the ws is
    for phi
        for r
            for z
    """

    prop_defaults = {
        "RMin": w7x.Defaults.CylindricalGrid.RMin,
        "RMax": w7x.Defaults.CylindricalGrid.RMax,
        "ZMin": w7x.Defaults.CylindricalGrid.ZMin,
        "ZMax": w7x.Defaults.CylindricalGrid.ZMax,
        "numR": w7x.Defaults.CylindricalGrid.numR,
        "numZ": w7x.Defaults.CylindricalGrid.numZ,
        "numPhi": w7x.Defaults.CylindricalGrid.numPhi,
        "PhiMin": w7x.Defaults.CylindricalGrid.PhiMin,
        "PhiMax": w7x.Defaults.CylindricalGrid.PhiMax,
    }

    def spacing(self):
        return self.numR, self.numPhi, self.numZ


class CartesianGrid(Base):
    """
    Order of grid in the ws is
    for phi
        for r
            for z
    """

    prop_defaults = {
        "numX": w7x.Defaults.CartesianGrid.numX,
        "numY": w7x.Defaults.CartesianGrid.numY,
        "numZ": w7x.Defaults.CartesianGrid.numZ,
        "ZMin": w7x.Defaults.CartesianGrid.ZMin,
        "ZMax": w7x.Defaults.CartesianGrid.ZMax,
        "XMin": w7x.Defaults.CartesianGrid.XMin,
        "XMax": w7x.Defaults.CartesianGrid.XMax,
        "YMin": w7x.Defaults.CartesianGrid.YMin,
        "YMax": w7x.Defaults.CartesianGrid.YMax,
    }


class Grid(Base):
    """
    see http://webservices.ipp-hgw.mpg.de/docs/fieldlinetracer.html#Grid
    Examples:
        >>> from w7x.flt import CylindricalGrid, Grid
        >>> cyl = CylindricalGrid(numPhi=49)
        >>> afsFilePath = "/fieldline/w7x/field_mfbe181x181x96.w7x.1000_1000_1000_1000_+0000_+0000.dboe_01.v_00_pres_00_it_12.dat"  # noqa
        >>> grid = Grid(cylindrical=cyl, afsFileName=afsFilePath)
        >>> grid.as_input()  # noqa
        (Grid){
            cylindrical = (CylindricalGrid){
                              RMin = 4.05
                              RMax = 6.75
                              ZMin = -1.35
                              ZMax = 1.35
                              numR = 181
                              numZ = 181
                              PhiMin = None (float)
                              PhiMax = None (float)
                              numPhi = 49
                          }
            hybrid = None (CylindricalGrid)
            afsFileName = /fieldline/w7x/field_mfbe181x181x96.w7x.1000_1000_1000_1000_+0000_+0000.dboe_01.v_00_pres_00_it_12.dat
            gridField = None (Points3D)
            fieldSymmetry = 5
        }

    """

    prop_defaults = {
        "cylindrical": None,
        "hybrid": None,
        "afsFileName": None,
        "gridField": None,
        "fieldSymmetry": w7x.Defaults.fieldSymmetry,
    }

    def __init__(self, **kwargs):
        super(Grid, self).__init__(**kwargs)
        """
        Do not use the hybrid unless you know what you are doing
        """
        if self.hybrid is not None:
            raise AttributeError("You should not use the hybrid grid.")
        if self.cylindrical is None:
            self.cylindrical = CylindricalGrid()


class Machine(Base):
    """
    Object to closely lie above web service Machine object
    Examples:
        >>> from w7x.flt import Machine, CartesianGrid
        >>> a = Machine(meshedModelsIds=w7x.MeshedModelsIds.divertor,
        ...             grid=CartesianGrid())
        >>> a.as_input()
        (Machine){
            meshedModels[] = [
        <BLANKLINE>
                              ]
            meshedModelsIds[] = [
                                 165,
                                 166,
                                 167,
                                 168,
                                 169
                                 ]
            assemblyIds[] = None (int)
            grid = (CartesianGrid){
                       XMin = -7
                       XMax = 7
                       YMin = -7
                       YMax = 7
                       ZMin = -1.5
                       ZMax = 1.5
                       numX = 500
                       numY = 500
                       numZ = 100
                   }
        }

    """

    prop_defaults = {
        "meshedModels": None,
        "meshedModelsIds": None,
        "assemblyIds": None,
        "grid": None,
    }
    ws_class_args = [1]

    def __init__(self, *args, **kwargs):
        """
        Args:
            *args (int | Mesh3D | str):
                int -> mm_ids
                Mesh3D | str -> MeshedModel
                str -> component name
        """
        if len(args) == 1 and isinstance(args[0], list):
            args = args[0]

        if len(args) == 1 and w7x.is_w7x_instance(args[0], self.get_ws_class()):
            # copy constructor
            pass
        elif len(args) > 0:
            raise DeprecationWarning("args only supported for copy constructor")
            # meshed_models = kwargs.pop('meshedModels', [])
            # meshed_models_ids = kwargs.pop('meshedModelsIds', [])
            # mm_ids = []
            # for arg in args:
            #     if issubclass(type(arg), tfields.Mesh3D):
            #         meshed_models.append(MeshedModel(arg))
            #     elif isinstance(arg, string_types):
            #         meshed_models.append(tfields.Mesh3D.load(arg))
            #     elif isinstance(arg, int):
            #         mm_ids.append(arg)
            #     else:
            #         arg_type = type(arg)
            #         raise NotImplementedError("No valid argument type "
            #                                   "'{arg_type}'.".format(**locals()))

            # for mm_id in mm_ids:
            #     ws_mm_id, own_mm_id = interprete_mm_id(mm_id)
            #     if own_mm_id is None:
            #         meshed_models_ids.append(mm_id)
            #     else:
            #         meshed_models.append(MeshedModel.from_mm_id(mm_id))

            # kwargs['meshedModels'] = meshed_models
            # kwargs['meshedModelsIds'] = meshed_models_ids
            # kwargs['grid'] = kwargs.pop('grid', CartesianGrid())

        meshed_models = kwargs.pop("meshedModels", [])
        for i in range(len(meshed_models)):
            mm = meshed_models[i]
            if issubclass(type(mm), tfields.Mesh3D):
                meshed_models[i] = MeshedModel(mm)
            if not isinstance(meshed_models[i], MeshedModel):
                raise ValueError("meshed model of wrong type {0}".format(type(mm)))
        mm_ids_db = []
        mm_ids_user = []
        mm_ids = kwargs.pop("meshedModelsIds", [])
        for mm_id in mm_ids:
            ws_mm_id, own_mm_id = interprete_mm_id(mm_id)
            if own_mm_id is None:
                mm_ids_db.append(mm_id)
            else:
                mm_ids_user.append(mm_id)
        if len(meshed_models) != 0:
            if len(mm_ids_user) == 0:
                mm_ids_user = [1000 + i for i in range(len(meshed_models))]
            else:
                if len(mm_ids_user) != len(meshed_models):
                    raise ValueError("len(meshed_models) != len(mm_ids_user)")
        elif len(mm_ids_user) != 0:
            raise ValueError("len(meshed_models) != len(mm_ids_user)")

        kwargs["meshedModels"] = meshed_models
        kwargs["meshedModelsIds"] = mm_ids_db
        kwargs["grid"] = kwargs.pop("grid", CartesianGrid())
        super(Machine, self).__init__(**kwargs)
        self._mm_ids_user = mm_ids_user

    @classmethod
    def default(cls):
        return cls.from_mm_ids(*w7x.Defaults.Machine.mm_ids)

    @classmethod
    def from_mm_ids(cls, *mm_ids, **kwargs):
        grid = CartesianGrid()
        return cls(grid=grid, meshedModelsIds=w7x.GeoSet(mm_ids))

    def get_mm_ids(self, user_defined=True, pre_defined=True):
        """
        Args:
            user_defined (bool): return mm_ids, defined by yourself (mm_id >= 1000)?
            pre_defined (bool): return mm_ids, defined by ws?
        """
        if pre_defined:
            mm_ids_db = self.meshedModelsIds or []
        else:
            mm_ids_db = []
        if user_defined:
            mm_ids_user = self._mm_ids_user
        else:
            mm_ids_user = []
        return mm_ids_user + mm_ids_db
        # pre_defined_mm_ids = []
        # user_defined_mm_ids = []
        # for mm_id in self.meshedModelsIds:
        #     if mm_id < 1000:
        #         if pre_defined:
        #             pre_defined_mm_ids.append(mm_id)
        #     else:
        #         if user_defined:
        #             user_defined_mm_ids.append(mm_id)
        # return tfields.lib.util.flatten(user_defined_mm_ids + pre_defined_mm_ids)

    @property
    def mm_ids(self):
        """
        database IDs from ComponentsDB
        """
        return self.get_mm_ids()

    @mm_ids.setter
    def mm_ids(self, mm_ids):
        self.meshedModelsIds = mm_ids

    def meshed_models(self, user_defined=True, pre_defined=False, geoFileDir=None):
        """
        Args:
            user_defined (bool): return MeshedModels, defined by yourself (mm_id >= 1000)?
            pre_defined (bool): return MeshedModels, defined by ws?
        """
        meshed_models = []
        # go through user_defined mm_ids
        for mm_id in self.get_mm_ids(user_defined=user_defined, pre_defined=False):
            geoFilePath = os.path.join(geoFileDir, "mm_id_{0}.obj".format(mm_id))
            meshedModel = MeshedModel.load(geoFilePath).as_input()
            meshed_models.append(meshedModel)

        # go through pre_defined mm_ids
        compDBServer = w7x.get_server(w7x.Server.addr_components_db)
        for mm_id in self.get_mm_ids(user_defined=False, pre_defined=pre_defined):
            meshedModel = MeshedModel(compDBServer.service.getComponentData(mm_id)[0])
            meshed_models.append(meshedModel)
        return meshed_models

    def as_mesh3d_list(self, user_defined=True, pre_defined=True):
        return [
            m.as_mesh3d()
            for m in self.meshed_models(
                user_defined=user_defined, pre_defined=pre_defined
            )
        ]

    def plot_poincare(self, phi=0, **kwargs):
        inters = self.intersect_mesh_phi_planes([phi])
        artists = []
        for mm_id, geoPoints in zip(self.mm_ids, inters[0]):
            geoSet = w7x.getGeoSet(mm_id)
            color = geoSet.color if geoSet else w7x.Defaults.Poincare.geometry_color
            plotKwargs = {"color": color, "method": "plot", "lw": 1}
            plotKwargs.update(kwargs)
            artists.append(w7x.plot_poincare_surfaces(geoPoints, **plotKwargs))
        return artists

    def intersect_mesh_phi_planes(self, phi_list_rad):
        """Get poincare points for a set of mm_ids and phis
        with intersectMeshPhiPlane service from the Mesh server web service.

        Args:
            phi_list_rad (list of floats): list of phi in rad
        Returns:
            phi_container (list[list[Points3D for each vertex] for each mm_id] for each phi)
        Raises:

        Examples:
            >>> import w7x
            >>> machine = w7x.flt.Machine.from_mm_ids(165, 166)
            >>> res = machine.intersect_mesh_phi_planes([0.0, 0.34, 6 * 0.34])
            >>> assert res[0][0][0].equal([[ 5.74323774,  0.,  0.98876001],
            ...                            [ 5.7433343 ,  0.,  0.98801955]],
            ...                           atol=1e-6, rtol=1e-6)
            >>> assert res[0][0][1].equal([[ 5.74201939,  0.,  0.99807376],
            ...                            [ 5.74099873,  0.,  1.00585009]],
            ...                           atol=1e-6, rtol=1e-6)

            >>> res = machine.intersect_mesh_phi_planes([0.0,  5*0.31])
            >>> assert res[0][0][0].equal([[5.74323774, 0., 0.98876001],
            ...                            [5.7433343 , 0., 0.98801955]],
            ...                           atol=1e-6, rtol=1e-6)
            >>> assert res[0][0][1].equal([[ 5.74201939, 0.,  0.99807376],
            ...                            [ 5.74099873, 0.,  1.00585009]],
            ...                           atol=1e-6, rtol=1e-6)
            >>> assert res[1][1][0].equal([[ 6.07343869, 1.55, -0.68684114],
            ...                            [ 6.07343617, 1.55, -0.68893482]],
            ...                           atol=1e-6, rtol=1e-6)
            >>> assert res[1][1][1].equal([[ 6.07340735, 1.55, -0.70596002],
            ...                            [ 6.07343617, 1.55, -0.68893482]],
            ...                           atol=1e-6, rtol=1e-6)
            >>> len(machine.mm_ids) == len(res[0])
            True

        """
        # Check input
        if not type(phi_list_rad) is list:
            raise TypeError(
                "phi_list_rad has to be of type list but is of type {}".format(
                    type(phi_list_rad)
                )
            )

        if len(phi_list_rad) == 0:
            return []

        mesh_server = w7x.get_server(w7x.Server.addr_mesh_server)

        # Run intersectMeshPhiPlane service for each phi AND mm_id
        phi_container = []
        for iPhi, phi in rna.log.progressbar(
            zip(range(len(phi_list_rad)), phi_list_rad)
        ):
            mm_id_container = []
            for mm_id in self.mm_ids:
                poincare_points_list = []
                # Set up mesh set
                mesh_set = mesh_server.types.SurfaceMeshSet()
                wrap = mesh_server.types.SurfaceMeshWrap()
                reference = mesh_server.types.DataReference()
                reference.dataId = "10"
                wrap.reference = reference
                mesh_set.meshes = [
                    wrap,
                ]

                # append mm_id_wrap
                mm_id_wrap = mesh_server.types.SurfaceMeshWrap()
                mm_id_reference = mesh_server.types.DataReference()
                mm_id_reference.dataId = str(mm_id)
                mm_id_wrap.reference = mm_id_reference
                mesh_set.meshes.append(mm_id_wrap)

                # ask web service for result and process it to right format
                # careful: Returns None if there is no component.
                res = mesh_server.service.intersectMeshPhiPlane(phi, mesh_set)

                if type(res) is list:
                    if res[0] is None:
                        return []
                    elif (
                        str(type(res[0]))
                        == "<class 'osa.xmltypes.PolygonPlaneIntersection'>"
                    ):
                        # Result is from MeshPhiIntersection
                        for intersection in res:
                            # res.surfs has entries for every phi
                            vertex_points = Points3D(
                                intersection.vertices, coord_sys=tfields.bases.CARTESIAN
                            )
                            vertex_points.transform(tfields.bases.CYLINDER)
                            # phi is correct with rounding precision before.
                            # This way it is perfectly correct
                            vertex_points[:, 1].fill(phi)
                            poincare_points_list.append(vertex_points)
                    else:
                        logging.error("Can not handle result list content.")
                elif res is None:
                    logging.debug(
                        "Result was None. Probably there was no"
                        "intersection of the mesh with this plane."
                    )
                else:
                    logging.error("Result is not of the right type")

                mm_id_container.append(poincare_points_list)
            phi_container.append(mm_id_container)
        return phi_container


class MagneticConfig(Base):
    """
    Object to closely lie above web service MagneticConfig object
    Examples:
        >>> import w7x
        >>> datFileName=w7x.Defaults.Paths.test_dat_file
        >>> cyl = w7x.flt.CylindricalGrid(numR=60, numPhi=101, numZ=10)
        >>> grid = w7x.flt.Grid(cylindrical=cyl)
        >>> m = w7x.flt.MagneticConfig.from_dat_file(datFileName, grid=grid)
        >>> flsType = m.as_input()
        >>> flsType  # noqa doctest: +ELLIPSIS
        (MagneticConfig){
            coils[] = None (PolygonFilament)
            coilsCurrents[] = None (float)
            coilsIds[] = None (int)
            coilsIdsCurrents[] = None (float)
            configIds[] = None (int)
            grid = (Grid){
                       cylindrical = (CylindricalGrid){
                                         RMin = 4.05
                                         RMax = 6.75
                                         ZMin = -1.35
                                         ZMax = 1.35
                                         numR = 60
                                         numZ = 10
                                         PhiMin = None (float)
                                         PhiMax = None (float)
                                         numPhi = 101
                                     }
                       hybrid = None (CylindricalGrid)
                       afsFileName = /fieldline/w7x/field_mfbe181x181x96.w7x.1000_1000_1000_1000_+0000_+0000.dboe_01.v_00_pres_00_it_12.dat
                       gridField = None (Points3D)
                       fieldSymmetry = 5
                   }
            inverseField = None (bool)
        }

    """

    prop_defaults = {
        "coils": None,
        "coilsCurrents": None,
        "coilsIds": None,
        "coilsIdsCurrents": None,
        "configIds": None,
        "grid": None,
        "inverseField": None,
    }

    @classmethod
    def default(cls):
        """
        Standard case with ideal coils
        """
        return cls.from_currents()

    @classmethod
    def from_id(cls, config_id, **kwargs):
        kwargs["configIds"] = [config_id]
        return cls(**kwargs)

    @classmethod
    def from_currents(cls, *currents, **kwargs):
        """
        Factory method for vacuum magnetic field construction with biot-savart.

        Args:
            *currents (float):
                first 5 currents -> non planar currents
                optional:
                    next 2 currents -> planar currents
                    next 2 | 10 currents -> sweep coil currents
                        If 2 currents are given, assume equal current drive
                        in all upper and lower sweep coils respectively
                    next 5 -> trim coils
            **kwargs:
                unit (str):
                    The currents given have the unit Aw like in the web service
                    input for magnetic configs (see coil_currents method
                    docstring)
                scale (float): For all relative units ('rw', 'r'), the 'scale'
                    attribute is active. The default is
                    w7x.Defaults.MagneticConfig.n_windings_npcs in 'rw'
                coilsIds: specify the coils by picking their ids from the
                    database.
                    NOTE: The order of the coils is not arbitrary. The coil ids
                        have to be given in the order:
                            1. 50 non planar coils NPC1-NPC5,NPC1-NPC5,...
                            2. 20 planar coils PCA,PCB,PCA,PCB,...
                            3. 10 sweep coils in the order of the half modules
                                (hm10, hm11,...)
                        Some coils_ids are already compiled into a list in
                        w7x.Defaults.MagneticConfig.coilsIds

        Examples:
            >>> import w7x

            Build a low iota configuration from relative winding currents
            >>> m = w7x.flt.MagneticConfig.from_currents(
            ...     1, 1, 1, 1, 1, -0.23, -0.23, 0.001, 0.001,
            ...     unit='rw')

            Build the same config with any other unit combination
            >>> units = ['rw', 'r', 'Aw', 'A']
            >>> for unit in units:
            ...     m_tmp = w7x.flt.MagneticConfig.from_currents(
            ...         *m.coil_currents(unit=unit),
            ...         unit=unit, scale=None)

            or with pre-defined configurations
            >>> m_alternative = w7x.flt.MagneticConfig.from_currents(
            ...     *w7x.config.Defaults.MagneticConfig.low_iota_rw,
            ...     unit='rw')

        """
        unit = kwargs.pop("unit", "rw")
        scale = kwargs.pop("scale", None)
        if scale is None:
            if unit == "rw":
                scale = w7x.Defaults.MagneticConfig.scale
            elif unit == "r" and scale is None:
                scale = (
                    w7x.Defaults.MagneticConfig.scale
                    / w7x.Defaults.MagneticConfig.n_windings_npcs
                )
            else:
                scale = 1.0
        if "r" not in unit and scale != 1:
            raise ValueError("Scale given for absolute currents.")
        if "coilsIdsCurrents" in kwargs:
            raise ValueError("Conflict between coilsIdsCurrents and currents")
        kwargs.setdefault("coilsIds", w7x.Defaults.MagneticConfig.coilsIds)

        if len(currents) < 14:
            if len(currents) < 5:
                if unit == "rw":
                    currents = [1.0] * 5  # standard case
                else:
                    raise ValueError(
                        "For all units except 'rw' we expect"
                        " at least the planar coil currents"
                    )
            # TODO The following is dangerous in case default is changed
            currents = list(currents)
            currents = currents + [0.0] * (14 - len(currents))

        """
        convert to Aw
        """
        currents = np.array(currents, dtype=float)
        currents = currents * scale

        """
        Transform symmetric currents
        """
        currents = list(currents)

        # only partial currents given
        if len(currents) in (5, 7, 9):
            currents = currents + [0.0] * (14 - len(currents))
        plcs = currents[:5] * 10  # non planar coils
        npcs = currents[5:7] * 10  # planar coils
        if len(currents) == 14:
            # stellarator symmetric sweep coils
            scs = currents[7:9] * 5  # sweep coils
            tcs = currents[9:]  # trim coils
        elif len(currents) == 22:
            # 10 sweep, 2 trim coils
            scs = currents[7:17]  # sweep coils
            tcs = currents[17:]  # trim coils
        else:
            raise ValueError(len(currents))

        currents = plcs + npcs + scs + tcs

        n_total_expected = 85  # yes, that looks hardcoded but is a constant
        if not len(kwargs["coilsIds"]) == len(currents) == n_total_expected:
            raise ValueError(
                "#Coils: {0}, \t#Currents: {1}, \texpected: {2}".format(
                    len(kwargs["coilsIds"]), len(currents), n_total_expected
                )
            )

        """
        Multiply by winding number for 'winding' units
        """
        if "w" not in unit:
            currents = np.array(currents)
            currents[:50] *= w7x.Defaults.MagneticConfig.n_windings_npcs
            currents[50:70] *= w7x.Defaults.MagneticConfig.n_windings_plcs
            currents[70:80] *= w7x.Defaults.MagneticConfig.n_windings_scs
            currents[80] *= w7x.Defaults.MagneticConfig.n_windings_tcs_a
            currents[81:82] *= w7x.Defaults.MagneticConfig.n_windings_tcs_b
            currents[82:] *= w7x.Defaults.MagneticConfig.n_windings_tcs_a
            currents = list(currents)

        kwargs["coilsIdsCurrents"] = currents

        # NOTE: tests if removing coilsIds with current 0 would speed things up
        #   were performed. Preliminary outcome: if at all a minor effect.
        kwargs["grid"] = kwargs.pop("grid", Grid())
        return cls(**kwargs)

    @classmethod
    def from_afs_file(cls, file_path, **kwargs):
        """
        set AFS BField file
        Args:
            file_path (str): afs file path
            **kwargs: forwarded to constructor of cls

        """
        if "grid" not in kwargs:
            raise AttributeError("Please specify a grid.")
        # create Grid with afsFileName and hybrid grid if existing.
        kwargs["grid"] = kwargs.pop("grid", Grid())
        # check that file is afs file
        if file_path.startswith(w7x.Defaults.Paths.tracer_network_share):
            file_path = file_path.lstrip(w7x.Defaults.Paths.tracer_network_share)
        if not file_path.startswith("/"):
            file_path = "/" + file_path
        if not file_path.startswith(
            "/" + w7x.Defaults.Paths.tracer_network_share_relative_path.split("/")[0]
        ):
            raise TypeError("File must be shared on network drive.")
        kwargs["grid"].afsFileName = file_path

        return cls(**kwargs)

    @classmethod
    def from_dat_file(
        cls,
        file_path_dat,
        network_share=w7x.Defaults.Paths.tracer_network_share,
        network_share_mount=w7x.Defaults.Paths.tracer_network_share_mount,  # NOQA
        network_share_workgroup=w7x.Defaults.Paths.tracer_network_share_workgroup,  # NOQA
        network_share_relative_path=w7x.Defaults.Paths.tracer_network_share_relative_path,  # NOQA
        **kwargs
    ):
        """
        Copy file to afs and return the instance with afsFile
        Args:
            see from_afs_file and __init__
        Examples:
            >>> import w7x
            >>> datFileName=w7x.Defaults.Paths.test_dat_file
            >>> cyl = w7x.flt.CylindricalGrid(numR=60, numPhi=101, numZ=10)
            >>> grid = w7x.flt.Grid(cylindrical=cyl)
            >>> m = w7x.flt.MagneticConfig.from_dat_file(datFileName, grid=grid)

        """
        file_path_dat = rna.path.resolve(file_path_dat)
        if file_path_dat.startswith(network_share):
            logging.warning(
                "Passing network share path to 'from_dat_file'."
                " Consider using 'from_afs_file'"
            )
            network_file_path_dat = file_path_dat
        else:
            network_file_path_dat = rna.path.resolve(
                network_share,
                network_share_relative_path,
                os.path.basename(file_path_dat),
            )

        # locally mounted?
        local_mount_file_path_dat = ""
        if os.name == "nt":
            local_mount_file_path_dat = network_file_path_dat
        if os.name == "posix":
            if os.path.exists(network_share_mount):
                local_mount_file_path_dat = rna.path.resolve(
                    network_share_mount,
                    network_share_relative_path,
                    os.path.basename(file_path_dat),
                )

        # copy if not existing
        if os.path.exists(local_mount_file_path_dat):
            # If file is on nas you have nothing to do
            file_already_exists_message = (
                "File {network_file_path_dat} is already existing. "
                "It is mounted as {local_mount_file_path_dat} ."
                "I will not copy."
            ).format(**locals())
            logging.info(file_already_exists_message)
        elif local_mount_file_path_dat:
            # Copy file to nas
            logging.info(
                "Copy {file_path_dat} to {local_mount_file_path_dat}".format(**locals())
            )
            rna.path.cp(file_path_dat, local_mount_file_path_dat)
        elif os.name == "posix":
            # Copy file to nas with smbclient.
            # Inspired by smbclient python package
            import subprocess

            logging.info(
                "Copy {file_path_dat} to {network_file_path_dat}".format(**locals())
            )
            remote_path = os.path.join(
                network_share_relative_path, os.path.basename(file_path_dat)
            )
            remote_path = remote_path.replace("/", "\\")  # nt compatible
            smbclient_cmd = [
                "smbclient",
                network_share.rstrip("/"),
                "-k",  # kerberos!
            ]
            smbclient_cmd = [x.encode("utf8") for x in smbclient_cmd]
            command_list = ["put"]
            command_list.extend(u'"%s"' % arg for arg in [file_path_dat, remote_path])
            command = u" ".join(command_list)
            cmd = smbclient_cmd + [b"-c", command.encode("utf8")]

            # retry because sambaclient is buggy
            n_tries = 10
            attempts = 0
            while attempts < n_tries:
                p = subprocess.Popen(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                )
                result = p.communicate()[0].rstrip().decode("utf8")
                if p.returncode != 0:
                    logging.warning(
                        u"Attempting to copy by smbclient failed. "
                        u"Number of attemps: {0} (cmd: {1})".format(
                            attempts, (b" ".join(cmd).decode("utf8"))
                        )
                    )
                    attempts += 1
                else:
                    break

            else:
                raise OSError(
                    u"Error on %r: %r" % (b" ".join(cmd).decode("utf8"), result)
                )

        return cls.from_afs_file(network_file_path_dat, **kwargs)

    def scale_currents(self, scale, unit="Aw"):
        """
        Scale the currents with scale
        """
        if self.coilsIdsCurrents is None:
            raise ValueError("No coil currents set yet.")
        if unit == "Aw":
            self.coilsIdsCurrents = [c * scale for c in self.coilsIdsCurrents]
        else:
            raise NotImplementedError(
                "Unit {unit} not implemented yet.".format(**locals())
            )

    def set_nominal_current(self, nominalCurrent, unit="Aw"):
        """
        Scale the currents to reach the nomianlCurrent
        """
        self.scale_currents(nominalCurrent / self.coilsCurrents[0], unit=unit)

    def coil_currents(self, unit="rw"):
        """
        Retrieve the coil currents (coilsIdsCurrents) in units of choice. This
        method is the opposite to tfields.MagneticConfig.from_currents.

        Args:
            unit (str): what spec
                'A': the current that is actually applied to the coil set
                'Aw': in A * winding number - the current that is applied if
                    there would be just 1 winding. Non planar coils have 108
                    windings each, planar coils, 36 and control coils 8
                    windings each. The trim coils have 46 or 72 windings,
                    depending on the coil type.
                'rw': 'Aw' normalized by (i.e. relative to - hence 'r') the
                    first non planar coil
                'r': 'A' normalized by (i.e. relative to - hence 'r') the first
                    non planar coil

        Examples:
            >>> from w7x import MagneticConfig
            >>> import numpy as np

            >>> m = MagneticConfig.from_currents(
            ...     1, 1, 1, 1, 1, 0.23, 0.23, 0, 0,
            ...     scale = 15000 * 108)
            >>> assert np.array_equal(
            ...     [round(x, 2) for x in m.coil_currents('rw')],
            ...      [1.0, 1.0, 1.0, 1.0, 1.0, 0.23, 0.23] + [0.0] * 7)
            >>> assert np.array_equal(
            ...     [round(x, 2) for x in m.coil_currents('r')],
            ...     [1.0, 1.0, 1.0, 1.0, 1.0, 0.69, 0.69] + [0.0] * 7)
            >>> assert np.array_equal(
            ...     [round(x) for x in m.coil_currents('Aw')],
            ...     [1620000.0, 1620000.0, 1620000.0, 1620000.0, 1620000.0,
            ...      372600.0, 372600.0] + [0.0] * 7)
            >>> assert np.array_equal([round(x) for x in m.coil_currents('A')],
            ...                       [15000.0, 15000.0, 15000.0, 15000.0,
            ...                        15000.0, 10350.0, 10350.0] + [0.0] * 7)

        """
        if unit == "r":
            currents = np.array(self.coil_currents("A"))
            currents = currents * 1.0 / currents[0]
            return currents

        npcs = self.coilsIdsCurrents[:5]
        plcs = self.coilsIdsCurrents[50:52]
        scs = self.coilsIdsCurrents[70:80]
        tcs = self.coilsIdsCurrents[80:]
        npcs = np.array(npcs)
        plcs = np.array(plcs)
        scs = np.array(scs)
        tcs = np.array(tcs)
        I_n = npcs[0]
        if (
            len(set(scs[[i for i in range(10) if i % 2 == 0]])) == 1
            and len(set(scs[[i for i in range(10) if i % 2 == 1]])) == 1
        ):
            # stellarator symmetric
            scs = scs[:2]

        if unit == "Aw":
            return list(npcs) + list(plcs) + list(scs) + list(tcs)
        if unit == "A":
            return (
                list(npcs * 1.0 / w7x.Defaults.MagneticConfig.n_windings_npcs)
                + list(plcs * 1.0 / w7x.Defaults.MagneticConfig.n_windings_plcs)
                + list(scs * 1.0 / w7x.Defaults.MagneticConfig.n_windings_scs)
                + list(tcs[:1] * 1.0 / w7x.Defaults.MagneticConfig.n_windings_tcs_a)
                + list(tcs[1:2] * 1.0 / w7x.Defaults.MagneticConfig.n_windings_tcs_b)
                + list(tcs[2:] * 1.0 / w7x.Defaults.MagneticConfig.n_windings_tcs_a)
            )
        if unit == "rw":
            return (
                list(npcs * 1.0 / I_n)
                + list(plcs * 1.0 / I_n)
                + list(scs * 1.0 / I_n)
                + list(tcs * 1.0 / I_n)
            )

        raise NotImplementedError(unit)

    def geiger_string(self):
        """
        Examples:
            >>> import w7x
            >>> w7x.MagneticConfig.default().geiger_string()
            '1000_1000_1000_1000_+0000_+0000'

        """
        return "{0:-04d}_{1:-04d}_{2:-04d}_{3:-04d}_{4:0=+5d}_{5:0=+5d}".format(
            *[int(x * 1000) for x in self.coil_currents()[1:8]]
        )

    def poincare_in_phi_plane(
        self, phi_list_rad, seeds=None, numPoints=300, errors=None
    ):
        """
        Calculate poincare points flux surface wise with field line tracer
        service from the web service.

        Args:
            phi_list_rad (list of floats): list of phi in rad
            seeds (PoincareSeeds): seeds object for initial points for field
            lines. Each seed in seeds will give a surface
            numPoints (int, optional): number of toroidal revolutions per
                seed = number of points per fluxSurface.

        Returns:
            list of w7x.Points3D instances: poincareFluxSurfaces.
            Each fluxSurface
                        ->length: number of seeds
                        ->meaning: it represents a closed flux surface
                        ->phi0 is attached to the object as a stamp

        Raises:

        Examples:
            >>> from w7x.flt import MagneticConfig, Points3D
            >>> config = MagneticConfig.default()
            >>> res = config.poincare_in_phi_plane(
            ...     [0.0, 2.0],
            ...     Points3D([[6.0, 0.0, 0.0]]), 3)
            >>> assert len(res) == 2
            >>> assert res[0].shape == (3, 3)
            >>> assert res[1].shape == (3, 3)
            >>> res2 = config.poincare_in_phi_plane(
            ...     [0.0, 2.0],
            ...     Points3D([[6.0, 0.0, 0.0]]), 1)
            >>> assert len(res2) == 2
            >>> assert res2[0].shape == (1, 3)
            >>> assert res2[1].shape == (1, 3)
            >>> res3 = config.poincare_in_phi_plane(
            ...     [2.0], Points3D([[6.0, 0.0, 0.0]]), 1)
            >>> assert len(res3) == 1
            >>> assert res3[0].shape == (1, 3)

        """
        if seeds is None:
            logging.info("No seeds given. I will build default seeds.")
            seeds = w7x.Defaults.Poincare.seeds

        seeds = Points3D(seeds)
        errors = errors or []

        if not numPoints > 0:
            raise ValueError("numPoints must be greater than 0")

        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)

        # Poincare Task
        task = field_line_server.types.Task()
        # parameter controling the accurancy of the calculation.
        task.step = w7x.Defaults.Poincare.stepSize
        task.poincare = field_line_server.types.PoincareInPhiPlane()
        task.poincare.numPoints = numPoints
        # list of phi in radians or just one phi.
        task.poincare.phi0 = phi_list_rad

        poincareFluxSurfaces = []
        nSuccess = 0
        nFailed = 0
        logging.info("Running through Seeds. Tracing each seed for all given phi.")
        for i_seed in range(len(seeds)):
            seed_point = seeds[i_seed : i_seed + 1]
            res = w7x.run_service(
                field_line_server.service.trace,
                seed_point.as_input(),
                self.as_input(),
                task,
                errors=errors,
            )

            if res is None:
                logging.warning("Turn {0} of {1}: fail".format(i_seed + 1, len(seeds)))
                nFailed += 1
                continue
            else:
                nSuccess += 1
                for surf in res.surfs:
                    p = Points3D(surf.points)
                    p.transform(tfields.bases.CYLINDER)
                    # phi is correct with rounding precision before. This way
                    # it is perfectly correct
                    p[:, 1].fill(surf.phi0)
                    poincareFluxSurfaces.append(p)
                logging.info("Turn %i of %i : success" % (i_seed + 1, len(seeds)))
        if nFailed > 0:
            logging.warning(
                "Trace finished with {0} succeded and {1} failed.".format(
                    nSuccess, nFailed
                )
            )
        else:
            logging.info(
                "Trace finished with {0} succeded and {1} failed.".format(
                    nSuccess, nFailed
                )
            )

        return poincareFluxSurfaces

    def magnetic_characteristics(self, points3D, taskStepSize=0.2, return_type=list):
        """
        Args:
            points3D (field_line_server.types.Points3D): points to retrieve
                characteristics from. Give None to take lcfs point
            taskStepSize (float)
            return_type (type): return either list or tfields.TensorFields.

        Examples:
            >>> from w7x.flt import MagneticConfig, Points3D
            >>> config = MagneticConfig.default()
            >>> mchars = config.magnetic_characteristics(
            ...     Points3D([[6.2, 0., 0.]]))  # This is long lasting.
            >>> assert len(mchars) == 1
            >>> mchar = mchars[0]
            >>> assert mchar.iota < 1
            >>> assert mchar.iota > 0.95
            >>> assert mchar.reff < 0.6
            >>> assert mchar.reff > 0.5
            >>> assert mchar.phi0 == 0.0

        """
        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)
        # process input points3D
        points3D = Points3D(points3D)
        points3D.transform(tfields.bases.CARTESIAN)

        # define task
        task = field_line_server.types.Task()
        task.step = taskStepSize
        task.characteristics = field_line_server.types.MagneticCharacteristics()  # NOQA
        task.characteristics.axisSettings = (
            field_line_server.types.AxisSettings()
        )  # NOQA

        # run web service
        logging.info("Retrieving MagneticCharacteristics with the points given.")
        # the None after 'task' can be machine boundary.
        result = w7x.run_service(
            field_line_server.service.trace,
            points3D.as_input(),
            self.as_input(),
            task,
            None,
            None,
        )

        if result is None:
            return None
        if return_type is list:
            return result.characteristics
        elif issubclass(return_type, tfields.TensorFields):
            fields = [[], [], [], [], [], []]
            for x in result.characteristics:
                fields[0].append(x.iota)
                fields[1].append(x.diota)
                fields[2].append(x.reff)
                fields[3].append(x.dreff)
                fields[4].append(x.phi0)
                fields[5].append(x.theta0)
            return return_type(points3D, *fields)

    def iota(self, points3D, taskStepSize=0.2, return_type=list):
        """
        Args: see magnetic_characteristics
        Examples:
            >>> from w7x.flt import MagneticConfig, Points3D
            >>> import tfields
            >>> config = MagneticConfig.default()

            # >>> iotas = config.iota(Points3D([[6.2, 0., 0.]]))
            # >>> assert iotas[0] > 0.95
            # >>> assert iotas[0] < 1.0

            >>> tf = config.iota(Points3D([[ 6.2, 0., 0.]]),
            ...                  return_type=tfields.TensorFields)
            >>> assert tf.equal([[6.2, 0. , 0. ]])
            >>> assert tf.fields[0].equal([0.9633], atol=1e-4)

        """
        mag_char = self.magnetic_characteristics(points3D, taskStepSize, return_type)
        if return_type is list:
            return [x.iota for x in mag_char]
        elif issubclass(return_type, tfields.TensorFields):
            mag_char.fields = mag_char.fields[0:1]
            return mag_char

    def find_axis_at_phi(
        self, phi=0, size=w7x.Defaults.Poincare.stepSize, settings=None
    ):
        """
        Args:
            phi (float): phi in rad
            size (float): step size
            settings (field_line_server.types.AxisSettings())
        Returns:
            Points3D: magnetic axis position at phi=<phi>
        """
        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)
        if settings is not None:
            settings = settings.as_input()
        else:
            settings = field_line_server.types.AxisSettings(1)
        logging.getLogger().info("Finding axis at phi.")
        result = w7x.run_service(
            field_line_server.service.findAxisAtPhi,
            phi,
            size,
            self.as_input(),
            settings,
        )
        return Points3D(result.points, coord_sys=tfields.bases.CARTESIAN)

    def magnetic_field(self, points3D):
        # assure points3D has the correct format and coord_sys
        points3D = Points3D(points3D)
        points3D.transform(tfields.bases.CARTESIAN)

        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)
        logging.getLogger().info("Evaluating magnetic field at points.")
        result = w7x.run_service(
            field_line_server.service.magneticField,
            points3D.as_input(),
            self.as_input(),
        )
        return Points3D(result.field, coord_sys=tfields.bases.CARTESIAN)

    def plot_poincare(self, phi=0, seeds=None, numPoints=300, errors=None, **kwargs):
        """
        Plot a poincare plot
        Args:
            phi (float): phi in rad
            seeds (Points3D): seed points. If None, take default seeds
        """
        surfs = self.poincare_in_phi_plane(
            [phi], seeds=seeds, numPoints=numPoints, errors=errors
        )
        return w7x.plot_poincare_surfaces(surfs, **kwargs)


class Run(object):
    """
    Container class that knows everything important about the w7x setup you want to use.
    Note:

    Args:
        magnetic_config (MagneticConfig):
        machine (Machine):
    Attributes:

    Examples:

        >>> from w7x.flt import Run
        >>> config = Run()

    """

    def __init__(self, magnetic_config=None, machine=None):
        self.magnetic_config = magnetic_config
        self.machine = machine

    @property
    def magnetic_config(self):
        """
        (w7x.flt.MagneticConfig): magnetic configuration of the run
        """
        return self._magnetic_config

    @magnetic_config.setter
    def magnetic_config(self, magnetic_config):
        if magnetic_config is None:
            magnetic_config = MagneticConfig.default()
        else:
            magnetic_config = MagneticConfig(magnetic_config)
        self._magnetic_config = magnetic_config

    @property
    def machine(self):
        """
        (w7x.flt.Machine): machine geometry of the run
        """
        return self._machine

    @machine.setter
    def machine(self, machine):
        if machine is None:
            machine = Machine.default()
        else:
            Machine(machine)
        self._machine = machine

    def find_lcfs(self, step=0.001, settings=None, max_time=None):
        """
        Examples:
            >>> import w7x
            >>> run = w7x.flt.Run()
            >>> lcfs_point = run.find_lcfs(0.1)  # This is long lasting.
            >>> assert lcfs_point[0, 0] > 6.2
            >>> assert lcfs_point[0, 0] < 6.21

        """
        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)
        if settings is None:
            settings = field_line_server.types.LCFSSettings(1)
        else:
            settings = settings.as_input()

        logging.getLogger().info("Retrieving Point on Last closed flux surface")
        lCFSPoint = w7x.run_service(
            field_line_server.service.findLCFS,
            step,
            self.magnetic_config.as_input(),
            self.machine.as_input(),
            settings,
            max_time=max_time,
        )  # find last closed flux surface
        return Points3D([[lCFSPoint.x, lCFSPoint.y, lCFSPoint.z]])

    def line_tracing(self, points, step=0.01, numSteps=300):
        """
        Args:
            points (w7x.flt.Points3D)
            step (flota): step width of the tracer task
        """
        points = Points3D(points)
        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)

        config = self.magnetic_config.as_input()

        task = field_line_server.types.Task()
        task.step = step
        task.lines = field_line_server.types.LineTracing()
        task.lines.numSteps = numSteps

        logging.info("starting line tracing ...")
        res = field_line_server.service.trace(points.as_input(), config, task, None)
        logging.info("... done")
        return [Points3D(line.vertices) for line in res.lines]

    def line_phi_span(self, points, phi, step=0.01):
        """
        Args:
            points (w7x.flt.Points3D)
            phi (float): phi in radian
            step (flota): step width of the tracer task
        """
        points = Points3D(points)
        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)

        config = self.magnetic_config.as_input()

        task = field_line_server.types.Task()
        task.step = step
        task.linesPhi = field_line_server.types.LinePhiSpan()
        task.linesPhi.phi = phi

        logging.info("starting line tracing ...")
        res = field_line_server.service.trace(
            points.as_input(), config, task, self.machine.as_input()
        )
        logging.info("... done")
        return [Points3D(line.vertices) for line in res.lines]

    def line_phi(self, points, phi, phi_tolerance=1.0 / 180 * np.pi, step=0.01):
        """
        Get the crossection of the field lines starting at <points> at exactly one
        phi (exact within the tolerance given by <phi_tolerance>).
        You can specify the direction of tracing by switching the bool
        self.magnetic_config.inverseField
        Args:
            points (w7x.Points3D): starting points of tracing
            phi (float): phi in radian
        Returns:
            w7x.Points3D: piercing points of the lines,
                starting at <points> at exactly one
                phi (exact within the tolerance given by <phi_tolerance>).
                The return value will be empty an empty seuqence of Points3D,
                if the step size was not small enough.
        """
        lines = self.line_phi_span(points, phi, step=step)
        container = []
        for line in lines:
            line.transform(tfields.bases.CYLINDER)
            inds = line.where_phi_between(phi - phi_tolerance, phi + phi_tolerance)
            line = line[inds[0][-1:]]
            line[:, 1] = phi
            container.append(line)
        return Points3D.merged(*container)

    def line_diffusion(
        self,
        start_points=w7x.Defaults.Diffusion.nRevolutions,
        diffusion=w7x.Defaults.Diffusion.diffusion,
        velocity=w7x.Defaults.Diffusion.velocity,
        size=w7x.Defaults.Diffusion.stepSize,
        mean_free_path=w7x.Defaults.Diffusion.meanFreePath,
        additiona_shifts=None,
        start_point_shift=w7x.Defaults.Diffusion.startPointShift,
        **kwargs
    ):
        """
        Args:
            start_points(int | str | Points3D):
                int: number of points to trace
                str: path to full start point set that should be traced
                Points3D: full start point set that should be traced
            diffusion (float): perp. diffusion coefficient
        Returns:
            tuple:
                w7x.flt.ConnectionLength:
                w7x.flt.ComponentLoad:
                w7x.flt.Points3D: start points lying on lcfs
        Examples:
            >>> import w7x
            >>> import tfields

            Triangle approximating the bean plane
            >>> mesh = tfields.Mesh3D([[5.4,0,-1], [5.4,0,1], [6.5,0,0]],
            ...                       faces=[[0,1,2]], coord_sys='cylinder')

            Machine constitued from bean plane triangle and divertor
            >>> machine = w7x.flt.Machine(meshedModels=[mesh],
            ...                           meshedModelsIds=[165])
            >>> run = w7x.flt.Run(machine=machine)

            Initial point starting very close to the triangle
            >>> initial_points = w7x.flt.Points3D([[6.0,0.0001,0]],
            ...                                   coord_sys='cylinder')
            >>> conn_len, comp_load, start_points = run.line_diffusion(
            ...     initial_points)
            >>> assert initial_points.equal(start_points,atol=1e-9)

            As expected, the triangle was hit only
            >>> conn_len.parts
            [0, 0]

            The mm_id 1000 is added automatically
            >>> conn_len.mm_ids
            [1000, 165]

        """
        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)
        if additiona_shifts is None:
            additiona_shifts = [0.00, 0.01, 0.01, 0.01, 0.01, 0.01, 0.1]

        nStart = None
        launchPointsPath = None
        if type(start_points) is int:
            nStart = start_points
            start_points = None
        elif type(start_points) is str:
            launchPointsPath = start_points
            start_points = None
        elif isinstance(start_points, tfields.Points3D):
            nStart = start_points.shape[0]
            if not isinstance(start_points, Points3D):
                start_points = Points3D(start_points)

        # define all setting inputs for diffusion
        fldTask = field_line_server.types.Task()
        fldTask.step = size
        line_diffusion = field_line_server.types.LineDiffusion()
        line_diffusion.diffusionCoeff = diffusion
        line_diffusion.freePath = mean_free_path
        line_diffusion.velocity = velocity
        fldTask.diffusion = line_diffusion
        fldTask.connection = field_line_server.types.ConnectionLength()
        fldTask.connection.limit = kwargs.pop("connectionLimit", 3000000.0)
        fldTask.connection.returnLoads = True
        config = self.magnetic_config.as_input()
        machine = self.machine.as_input()
        logging.info("Machine flsType: \n{machine}".format(**locals()))
        logging.info("Config flsType: \n{config}".format(**locals()))

        # check start_points validity
        if start_points is None and launchPointsPath is None:
            logging.info(
                "Creating random start lcfs_init_points on last closed flux surface"
            )
            lcfs_init_points = self.find_lcfs()

            """
            remove small value to be save to be inside
            offset changed on 2.11 from 0.001 to 0.02. It shows, that the
            computation time increase due to longer diffusion path
            is waaay less than repeating a whole shot (at least for D=1m^s/s).
            """
            lcfs_init_points[:, 0] -= 0.02 + start_point_shift

        for iShift, additionalShift in enumerate(additiona_shifts):
            if iShift == 0 and start_points is not None:
                # before start_points could have been set by fld it has been
                # given
                logging.info("Start points were given.")
            elif launchPointsPath is None:
                lcfs_init_points[:, 0] -= additionalShift
                logging.info(" Startpoint: {lcfs_init_points}".format(**locals()))

                traceTask = field_line_server.types.Task()
                traceTask.step = 0.5
                line = field_line_server.types.LineTracing()
                line.numSteps = nStart - 1
                traceTask.lines = line

                logging.info("Trace this point " + str(line.numSteps) + "-times.")
                res = w7x.run_service(
                    field_line_server.service.trace,
                    lcfs_init_points.as_input(),
                    config,
                    traceTask,
                    machine,
                    max_time=None,
                )

                start_points = res.lines[0].vertices

                # END  creating random start_points on last closed flux surface
            else:
                logging.info("Reading launch points")
                start_points = Points3D.load(launchPointsPath)

            # force start_points to CARTESIAN and check dimension
            start_points = w7x.Points3D(start_points)
            start_points.transform(tfields.bases.CARTESIAN)

            logging.info("Line diffusion forwards.")
            # max_time could be something like int(round(line.numSteps / 4000. * 2000)))
            resForward = w7x.run_service(
                field_line_server.service.trace,
                start_points.as_input(),
                config,
                fldTask,
                machine,
                max_time=None,
            )
            resForwardClass = ConnectionLength(resForward.connection)
            if len(resForwardClass) == nStart:
                logging.info("Length {} accepted!".format(len(resForwardClass)))
                if launchPointsPath is None:
                    start_points = Points3D(start_points)
                break
            elif launchPointsPath is not None:
                logging.warning(
                    "LaunchPointsPath is given but number of HitPoints"
                    "is not meeting requested number."
                )
                break
            elif iShift == len(additiona_shifts) - 1:
                logging.error(
                    "Also after shift of {} number of H".format(sum(additiona_shifts))
                )
                raise ValueError(
                    "Also after shift of %s number of H" % sum(additiona_shifts)
                )
            else:
                logging.warning(
                    "Restart the forward tracing with a new lCFSInitPoint "
                    "offset of {0} since the number of output points ({1}) "
                    "does not meet opts.nRevolutions({2})".format(
                        sum(additiona_shifts[: iShift + 2]),
                        len(resForwardClass),
                        nStart,
                    )
                )

        logging.info("Line diffusion backwards.")
        config.inverseField = True
        # max_time estimate: int(round(line.numSteps / 4000. * 2000)))
        resInvers = w7x.run_service(
            field_line_server.service.trace,
            start_points.as_input(),
            config,
            fldTask,
            machine,
            max_time=None,
        )

        connection_length = ConnectionLength(
            resForward.connection + resInvers.connection
        )
        connection_length.mm_ids = self.machine.mm_ids

        component_load = ComponentLoad(
            resForward.loads.components + resInvers.loads.components
        )
        component_load.mm_ids = self.machine.mm_ids
        return connection_length, component_load, start_points

    def connection_length(self, points, **kwargs):
        """
        Args:
            points: Points3D
            **kwargs:
                limit (float): lenght limit of line tracing
                step (float): step size of line tracing
                diffusion (field_line_server.type.LineDiffusion()): add diffusion
                    to the tracing

        Returns:
            ConnectionLength

        """
        step = kwargs.pop("step", 5e-3)
        limit = kwargs.pop("limit", 2.0e4)
        diffusion = kwargs.pop("diffusion", None)

        points.transform(tfields.bases.CARTESIAN)

        field_line_server = w7x.get_server(w7x.Server.addr_field_line_server)
        task = field_line_server.types.Task()
        task.step = step
        con = field_line_server.types.ConnectionLength()
        con.limit = limit
        con.returnLoads = False
        task.connection = con

        if diffusion:
            task.diffusion = diffusion

        logging.info("Start tracing ...")
        res = w7x.run_service(
            field_line_server.service.trace,
            points.as_input(),
            self.magnetic_config.as_input(),
            task,
            self.machine.as_input(),
        )
        connection_length = ConnectionLength(res.connection)
        connection_length.mm_ids = self.machine.mm_ids
        lengths = np.array(connection_length.lengths)
        lengths[lengths == limit] = np.inf
        connection_length.lengths = list(lengths)
        return connection_length

    def plot_poincare(self, phi, seeds=None, **kwargs):
        """
        forward to self.magnetic_config.plot_poincare and
        self.machine.plot_poincare
        """
        artists = []
        if self.magnetic_config is not None:
            artists.extend(
                self.magnetic_config.plot_poincare(phi, seeds=seeds, **kwargs)
            )
        if self.machine is not None:
            artists.extend(self.machine.plot_poincare(phi, **kwargs))
        return artists


if __name__ == "__main__":
    import doctest

    # doctest.run_docstring_examples(MagneticConfig.iota, globals())
    # quit()
    doctest.testmod()



"""
abstract classes to work with the webservice revolving around the osa package
"""
import w7x.lib
import six
import copy
from six import string_types
import osa
import logging
import numpy as np
import rna
import tfields
try:
    # For Python 3.0 and later
    from urllib.error import HTTPError
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import HTTPError


def compare_error_attributes(error, errors=None, attributes=None,
                             messageContent=None, **kwargs):
    """
    Annotation:
        if you want a content in the msg, use args attribute.
    Returns:
        True if error.attributes == error_list.attributes else False
    Examples:
        >>> from w7x import core
        >>> d = {'errors': [core.HTTPError(None, 500, None, None, None),
        ...                 core.HTTPError("ASDF", 600, None, None, None)],
        ...      'attributes': ['code'],
        ...      'messageContent': None}
        >>> core.compare_error_attributes(core.HTTPError(None, 500, None, None, None), **d)  # NOQA
        [True, False]
        >>> e = {'errors': [RuntimeError("A special Message content"),
        ...                 core.HTTPError("ASDF", 600, None, None, None)],
        ...      'attributes': ['args']}
        >>> core.compare_error_attributes(RuntimeError("This is A special Message content for you"), **e)  # NOQA
        [True, False]

    """
    errors = errors or []
    attributes = attributes or ['code']
    comparisons = []
    for inst in errors:
        return_bool = True
        if type(error) is not type(inst):  # must be same type ...
            return_bool = False
        else:
            for attr in attributes:  # ... and contain the same requested attribute values
                if not hasattr(error, attr) or not hasattr(inst, attr):
                    return_bool = False
                    break
                if attr == 'args':
                    if len(error.args) == 0:
                        raise TypeError("Length of args is not suiting.")
                    if not inst.args[0] in error.args[0]:
                        return_bool = False
                        break
                elif not getattr(error, attr) == getattr(inst, attr):
                    return_bool = False
                    break
        comparisons.append(return_bool)
    return comparisons


def run_service(fun, *args, **kwargs):
    """
    Run a service function with the arguments args. Check for HTTPError
    occurence and redo if you get it.
    Kwargs:
        max_tries (int): Maximum tries, Default is 1
        max_time (int): Maximum time for the process [s]. Default is inf (<None> value)
        errors (list of dicts describing errors): Each dict has the
            keywords:
                'errors' (list of Error instances)
                'attributes' (list of attributes)
                'action' (str): 'retry' / 'skip'
            Perform action if error.attributes == errors.attributes for any
            error
    Returns:
        Result of service or function. None if Error from errorDict says skip
    """
    log = logging.getLogger()
    max_tries = kwargs.pop("max_tries", 1)
    if max_tries < 1:
        raise ValueError("max_tries needs to be 1 at minimum.")
    max_time = kwargs.pop("max_time", None)
    errors_default = [{'errors': [HTTPError(None, 500, None, None, None)],
                       'attributes': ['code'],
                       'action': 'retry'},
                      {'errors': [HTTPError(None, 404, None, None, None)],
                       'attributes': ['code'],
                       'action': 'retry'},
                      {'errors': [HTTPError(None, 413, None, None, None)],
                       'attributes': ['code'],
                       'action': 'retry'},
                      {'errors': [RuntimeError('Grid_CylLowendel')],
                       'attributes': ['args'],
                       'action': 'skip'},
                      {'errors': [RuntimeError('ThreadTracer failed.')],
                       'attributes': ['args'],
                       'action': 'skip'},
                      {'errors': [RuntimeError('zero field')],
                       'attributes': ['args'],
                       'action': 'skip'}]
    errors = kwargs.pop('errors', [])
    errors.extend(errors_default)

    result = None
    try_count = 1

    @w7x.lib.timeout.timeout(max_time)
    def run(fun, *args):
        return fun(*args)

    while result is None:
        try:
            result = run(fun, *args)  # run service with arguments
            if result is None:
                log.warning("Result is really None")
            break
        except w7x.lib.timeout.TimeoutError:
            log.error("TimeoutError. Took more than %s seconds." % max_time)
            if try_count < max_tries:
                try_count += 1
                log.info("Retry service %s." % fun)
            else:
                raise  # Also allows result being <None>
        except Exception as err:
            log.error(err)
            skip = False
            retry = False
            for eD in errors:
                if any(compare_error_attributes(err, **eD)):
                    if eD['action'] == 'retry':
                        retry = True
                    elif eD['action'] == 'skip':
                        skip = True
                    else:
                        raise TypeError("%s is no allowed Action" % eD['action'])
            if skip:
                log.info("Skip service %s." % fun)
                break
            elif retry:
                if try_count < max_tries:
                    try_count += 1
                    log.info("Retry service %s." % fun)
                    continue
            raise
    return result


SERVERCACHE = {}


def get_server(address):
    """
    Cached method to retrieve osa.Client from adress.
    Timeout is implemented.
    Examples:
        TODO: this example only works if the server can connect
        >>> import w7x
        >>> addr = w7x.Server.addr_field_line_server
        >>> addr in str(w7x.get_server(addr))
        True

    """
    if address in SERVERCACHE:
        return SERVERCACHE[address]

    server = None
    log = logging.getLogger()

    server = run_service(osa.Client, address, max_tries=1, max_time=1)
    try:
        server = run_service(osa.Client, address, max_tries=1, max_time=1)
    except Exception:
        log.error("Server at {0} could not connect.".format(address))
    SERVERCACHE[address] = server
    return server


def get_ws_class(ws_server, ws_class):
    """
    Args:
        ws_server (str): address of webservice
        ws_class (str): name of webService type class
    Returns:
        server.types.ws_class
    """
    server = get_server(ws_server)
    return getattr(server.types, ws_class)


def is_w7x_instance(obj, tpe, convert=True):
    """
    Check, if an object is instance of a web server class
    Args:
        obj (object instance): instance to check the ...
        tpe (class or <webservice>.types.<class>): ... type for
        convert (bool): convert all obj and tpe to osa before checking
            you can provide conversion ability by defining the methods
            'as_input' for objs and 'get_ws_class' for tpes
    Returns:
        bool
    Examples:
        >>> import w7x
        >>> p = w7x.Points3D([[5, 1, 0]])

        checking osa type directly
        >>> fieldLineServer = w7x.get_server(w7x.Server.addr_field_line_server)
        >>> w7x.core.is_w7x_instance(p, fieldLineServer.types.Points3D)
        True
        >>> w7x.core.is_w7x_instance(p.as_input(), fieldLineServer.types.Points3D)
        True

        checking objects that provide get_ws_class methods
        >>> w7x.core.is_w7x_instance(p, w7x.Points3D)
        True
        >>> w7x.core.is_w7x_instance(p.as_input(), w7x.Points3D)
        True

        checking only osa types
        >>> w7x.core.is_w7x_instance(p, w7x.Points3D, convert=False)
        True
        >>> w7x.core.is_w7x_instance(p.as_input(), w7x.Points3D, convert=False)
        False
        >>> w7x.core.is_w7x_instance(p, fieldLineServer.types.Points3D, convert=False)
        False
        >>> w7x.core.is_w7x_instance(p.as_input(), fieldLineServer.types.Points3D, convert=False)
        True

        test the same for Base derived instance
        >>> m = w7x.MagneticConfig.default()
        >>> w7x.core.is_w7x_instance(m.as_input(), w7x.MagneticConfig)
        True
        >>> w7x.core.is_w7x_instance(m, w7x.MagneticConfig)
        True
        >>> w7x.core.is_w7x_instance(m, w7x.MagneticConfig.get_ws_class())
        True

    """
    # convert all Base derived objs and tpes to osa
    if convert:
        if issubclass(obj.__class__, Base) or hasattr(obj, 'as_input'):
            obj = obj.as_input()
        if issubclass(tpe, Base) or hasattr(tpe, 'get_ws_class'):
            tpe = tpe.get_ws_class()

    if isinstance(tpe, osa.xmltypes.ComplexTypeMeta):
        if convert or isinstance(obj.__class__, osa.xmltypes.ComplexTypeMeta):
            return obj.__class__.__name__ == tpe.__name__

    return isinstance(obj, tpe)


class Base(object):
    """
    set this to a dictionary with keys and defaults
    """
    prop_defaults = None
    prop_order = None

    ws_server = None
    ws_class = None
    ws_class_args = None
    ws_class_kwargs = None

    def __init__(self, *args, **kwargs):

        self.prop_defaults = self.prop_defaults or {}
        self.prop_order = self.prop_order or []

        if len(args) == 1 and is_w7x_instance(args[0], self.__class__):
            # copy constructor from cls or related ws class
            other = args[0]
            if is_w7x_instance(other, self.get_ws_class(), convert=False):
                for attr in self.prop_attrs():
                    if hasattr(other, attr):
                        kwargs[attr] = getattr(other, attr)
            else:
                kwargs = other.prop_dict(kwargs)
            if len(args) > 1:
                raise ValueError("More than one argument given in copy "
                                 "constructor.")
        else:
            # update kwargs with arguments defined in args
            for attr, arg in zip(self.prop_order, args):
                if attr in kwargs:
                    raise AttributeError("Attribute {attr} specified in args "
                                         "and kwargs! I will use args!"
                                         .format(**locals()))
                kwargs[attr] = arg

        # default properties
        props = self.prop_dict()
        props.update(self.prop_defaults)
        # set attributes from kwargs or props
        for key, default in six.iteritems(props):
            val = kwargs.pop(key, default)
            setattr(self, key, val)

        if len(kwargs) > 0:
            raise AttributeError('kwargs have unused arguments %s' % kwargs.keys())

    def prop_attrs(self):
        """
        Returns:
            list of str: properties that are occuring in the osa type input
        """
        try:
            attrs = dir(self.get_ws_class())
        except Exception:
            log = logging.getLogger()
            log.warning("could not connect to server and thus not dynamically "
                        "get the property defaults.")
            if self.prop_defaults is not None:
                attrs = self.prop_defaults.keys()

        prop_attrs = []
        for attr in attrs:
            if attr.startswith('_'):
                continue
            if attr in ['from_file', 'from_xml', 'to_file', 'to_xml']:
                continue
            prop_attrs.append(attr)
        return prop_attrs

    def prop_dict(self, kwargs=None):
        kwargs = kwargs or {}
        for key in self.prop_attrs():
            kwargs[key] = getattr(self, key, None)
        return kwargs

    def __deepcopy__(self, memo):
        """
        copy with the copy constructor
        """
        kwargs = copy.deepcopy(self.prop_dict(), memo)
        return self.__class__(**kwargs)

    def copy(self):
        """
        copy with deepcopy
        """
        return copy.deepcopy(self)

    @classmethod
    def get_ws_class(cls):
        """
        Returns the osa class version of this class
        """
        ws_class = cls.ws_class or cls.__name__
        return get_ws_class(cls.ws_server, ws_class)

    def to_file(self, path):
        """
        Forward to xml.to_file(self, path)
        Examples:
            >>> import w7x
            >>> mc = w7x.flt.MagneticConfig.default()
            >>> mc.to_file('/tmp/mc.xlm')
            >>> mc_loaded = w7x.flt.MagneticConfig.from_file('/tmp/mc.xlm')
            >>> mc.as_input() == mc_loaded.as_input()
            True

        """
        self.as_input().to_file(rna.path.resolve(path))

    @classmethod
    def from_file(cls, path):
        return cls(cls.get_ws_class().from_file(rna.path.resolve(path)))

    def as_input(self):
        """
        return copy in get_ws_class format. Chain this to the attributes.
        """
        cls = self.get_ws_class()
        ws_class_args = self.ws_class_args or []
        ws_class_kwargs = self.ws_class_kwargs or {}
        instance = cls(*ws_class_args, **ws_class_kwargs)
        for (prop, default) in six.iteritems(self.prop_defaults):
            value = self.__dict__.get(prop, default)
            if value is not None:
                if issubclass(value.__class__, Base) or hasattr(value, 'as_input'):
                    value = value.as_input()
                elif hasattr(value, '__iter__') and not isinstance(value,
                                                                   string_types):
                    value = [v.as_input() if hasattr(v, 'as_input')
                             else v for v in value]
                setattr(instance, prop, value)
        return instance


class BasePoints3D(tfields.Points3D, Base):
    """Imitation of the field_line_server Points3D Type.
    Inheriting from tfield.Points3D so the coordinate system is tracked and
    coordinate transformations are inherently possible

    Args:
        many ways to initialize:
        1.
            like tfields.Points3D
        2.
            points3D (osa.Points3D): copyConstructor
        3.
            pointsList (list): list of triples in varioues formats
    Attributes:

    Examples:
        see tfields.points3D.Points3D

        One further constructor implemented:
        >>> import w7x
        >>> ws_points = w7x.Points3D([[1, 2, 3], [4, 5, 6]])
        >>> ws_points_2 = w7x.Points3D(ws_points.as_input())
        >>> assert ws_points.equal(ws_points_2)

    """

    def __new__(cls, tensors, *args, **kwargs):
        if w7x.is_w7x_instance(tensors,
                               cls.get_ws_class(),
                               convert=False):
            tensors = np.array([tensors.x1, tensors.x2, tensors.x3]).T
        obj = tfields.Points3D.__new__(cls, tensors, *args, **kwargs)
        return obj

    def __init__(self, *args, **kwargs):
        for attr in self.__slots__:  # TODO remove that
            kwargs.pop(attr, None)
        super().__init__(self, *args, **kwargs)

    def as_input(self):
        """
        return field_line_server type copy of self
        """
        raw_points = self.get_ws_class()()
        raw_points.x1 = list(self[:, 0])
        raw_points.x2 = list(self[:, 1])
        raw_points.x3 = list(self[:, 2])
        return raw_points

    def to_segment_one(self, mirror_z=True):
        """
        Map the points to the first module of w7x and mirror to positive z
        if mirror_zOption is True. The mirror option is interesting for the
        divertor for example.
        Examples:
            >>> import w7x
            >>> import numpy as np
            >>> pStart = w7x.Points3D([[6, np.pi, 1],
            ...                        [6, np.pi / 5 * 3, -1]],
            ...                       coord_sys='cylinder')
            >>> pStart.to_segment_one()
            >>> pStart
            Points3D([[ 6.        , -0.62831853,  1.        ],
                      [ 6.        ,  0.62831853,  1.        ]])

        """
        with self.tmp_transform(tfields.bases.CYLINDER):
            offset_segment_0 = -2 * np.pi / 10
            self.to_segment(0, 5, 1, offset=offset_segment_0)
            if mirror_z:
                condition = self[:, 2] < 0
                self.mirror([1, 2], condition=condition)

    def where_phi_between(self, phi_min, phi_max):
        """
        Args:
            phi_min (float): minimum phi in radian
            phi_max (float): maximum phi in radian
        Returns:
            output of np.where with condition
        """
        with self.tmp_transform(tfields.bases.CYLINDER):
            phi = self[:, 1]
        if phi_min < phi_max:
            return np.where((phi_min <= phi) & (phi <= phi_max))
        elif phi_min > phi_max:
            return np.where(np.logical_not((phi_max < phi) & (phi < phi_min)))
        else:
            return np.where(phi == phi_min)

    def phi_between(self, phi_min, phi_max):
        """
        Args:
            phi_min (float): minimum phi in radian
            phi_max (float): maximum phi in radian
        Returns:
            bool: if phi of all points is in between the given values
        """
        return len(self.where_phi_between(phi_min, phi_max)[0]) == len(self)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import rna
import numpy as np
import matplotlib.pyplot as plt


def plot_tile_loads(loads, **kwargs):
    """
    Args:
        mask (np.ndarray): 2d array with mask for each value to plot wiht alpha=0
        text (str): annotation which is added above the picture
    """
    po = rna.plotting.PlotManager(kwargs)
    loads = np.array(loads)
    if len(loads.shape) == 1:
        raise DeprecationWarning()
    mask = np.array(po.pop("mask", None))
    text = po.pop("text", None)
    if mask is not None:
        cmap, vmin, vmax = po.get_norm_args(cmap_default="viridis")
        loads = rna.plotting.colors.to_colors(
            loads, fmt="rgba", cmap=cmap, vmin=vmin, vmax=vmax
        )
        alphas = np.ones(mask.shape)
        alphas[mask] = 0.0
        loads[["a"]] = alphas.reshape(loads.shape)
        # get rid of rgba dtype
        loads = np.lib.recfunctions.structured_to_unstructured(loads)
    artist = po.axes.imshow(loads, **po.plot_kwargs)
    po.axes.set_xticks(np.arange(-0.5, loads.shape[1] + 0.5, 1), minor=True)
    po.axes.set_yticks(np.arange(-0.5, loads.shape[0] + 0.5, 1), minor=True)

    plt.rc("grid", linestyle="solid", color="black")
    po.axes.grid(color="k", which="minor")

    po.axes.tick_params(
        axis="both",  # changes apply to both axis
        which="both",  # both major and minor ticks are affected
        bottom=False,  # ticks along the bottom edge are off
        top=False,  # ticks along the top edge are off
        labelbottom=False,
    )  # labels along the bottom edge are off
    po.axes.yaxis.set_ticks_position("none")
    plt.setp(po.axes.get_yticklabels(), visible=False)

    # optional: add an annotation
    if text is not None:
        po.axes.text(2, -2, text)
    return artist



#!/user/bin/env python
"""
w7x option starter
"""
import os
import numpy as np
import sys
import unittest
import doctest
import pathlib
import argparse

import tfields
import w7x


def run_doctests():
    """
    Find all doctests and execute them
    """
    this_dir = pathlib.Path(__file__).parent.resolve()
    for f in list(this_dir.glob("**/*.py")):
        doctest.testfile(str(f.resolve()), module_relative=False)


def load_unittests(loader=None, suite=None):
    if loader is None:
        loader = unittest.TestLoader()
    if suite is None:
        suite = unittest.TestSuite()
    parent = pathlib.Path(__file__).parent.parent
    for f in list(parent.glob("*/test*.py")):
        sys.path.insert(0, str(f.parent))
        mod = __import__(f.name[:-3])
        for test in loader.loadTestsFromModule(mod):
            suite.addTests(test)
        sys.path.remove(str(f.parent))
    return suite


def run_unittests(arg):
    run_doctests()
    # unittest.main(defaultTest='load_unittests')


def diffuse(args):
    """
    diffuse a extended vmec run and save the result
    """
    raise NotImplementedError("Paths should be set by user.")
    localDir = tfields.lib.in_out.resolve(
        "~/Data/EXTENDER/{args.vmec_id}/".format(**locals())
    )
    datPath = os.path.join(localDir, "{args.vmec_id}.dat".format(**locals()))

    cyl = w7x.extender.getHybridFromDat(datPath)
    grid = w7x.flt.Grid(cylinder=cyl)
    config = w7x.flt.MagneticConfig.from_dat_file(datPath, grid=grid)
    run = w7x.flt.Run(config)

    saveDir = os.path.join(args.baseDir, "{args.vmec_id}".format(**locals()))
    tfields.lib.in_out.mkdir(saveDir, isDir=True)
    path = tfields.lib.in_out.resolve(
        os.path.join(saveDir, "{args.vmec_id}-fld.nest.npz".format(**locals()))
    )
    tracerConnectionResults, componentLoads, startPoints3D = run.line_diffusion(
        startPoints=args.startPoints
    )


def poincare(args):
    """ poincare plot creation """
    phiList = args.phi.values

    if args.phi.deg:
        phiList = [val / 180 * np.pi for val in phiList]

    if not args.assemblies.off:
        machine = w7x.flt.Machine.from_mm_ids(*args.assemblies.values)
    else:
        machine = w7x.flt.Machine()

    relativeCurrents = args.magneticConfig.relativeCurrents
    datPath = args.magneticConfig.path
    if datPath:
        datPath = tfields.lib.in_out.resolve(datPath)
        cyl = w7x.extender.getHybridFromDat(datPath)
        grid = w7x.flt.Grid(cylinder=cyl)
        magneticConfig = w7x.flt.MagneticConfig.from_dat_file(datPath, grid=grid)
    elif relativeCurrents:
        magneticConfig = w7x.flt.MagneticConfig.createWithCurrents(
            relativeCurrents=relativeCurrents
        )
    else:
        magneticConfig = w7x.flt.MagneticConfig.createWithCurrents()

    """ plotting """
    axis = tfields.plotting.gca(2)
    for phi in phiList:
        axis.grid(color="lightgrey")
        machine.plot_poincare(phi, axis=axis)
        magneticConfig.plot_poincare(phi, axis=axis)
        tfields.plotting.save(
            "~/tmp/poincare-{phi:.4f}".format(**locals()), "png", "pgf", "pdf"
        )
        axis.clear()


if __name__ == "__main__":
    # create the top-level parser
    parser = argparse.ArgumentParser(prog="W7-X tools app")
    subparsers = parser.add_subparsers(help="sub-command help")

    # create the parser for the "test" command
    parserExtend = subparsers.add_parser("test", help="test help")
    parserExtend.set_defaults(func=run_unittests)

    # create the parser for the "extend" command
    parserExtend = subparsers.add_parser("extend", help="extend help")
    parserExtend.add_argument("vmec_id", type=str, help="vmec_id to extend")
    parserExtend.set_defaults(func=w7x.extender.extend)

    # create the parser for the "diffuse" command
    parserDiffuse = subparsers.add_parser("diffuse", help="diffuse help")
    parserDiffuse.add_argument("vmec_id", type=str, help="already extended vmec_id")
    parserDiffuse.add_argument(
        "--startPoints",
        type=int,
        help="hit points = 2 * " "startPoints (forward and backward).",
        default=12500,
    )
    parserDiffuse.add_argument(
        "--baseDir",
        type=str,
        default="~/Data/Strikeline/",
        help="already extended vmec_id",
    )
    parserDiffuse.set_defaults(func=diffuse)

    # create the parser for the "poincare" command
    parserPoincare = subparsers.add_parser("poincare", help="poincare help")
    parserPoincare.add_argument(
        "--baseDir",
        type=str,
        default="~/Data/Strikeline/",
        help="already extended vmec_id",
    )
    parserPoincare.add_argument(
        "--phi", dest="phi.values", nargs="*", type=float, default=[0.0]
    )
    parserPoincare.add_argument(
        "--phi.deg",
        dest="phi.deg",
        help="switch phi from radian to degree",
        action="store_true",
    )
    parserPoincare.add_argument(
        "--assemblies",
        dest="assemblies.values",
        nargs="+",
        type=str,
        default=w7x.Defaults.Machine.mm_ids,
    )
    parserPoincare.add_argument("--assemblies.off", action="store_true")
    parserPoincare.add_argument(
        "--magneticConfig.relativeCurrents",
        help="relative currents in case of vacuum config",
    )
    parserPoincare.add_argument(
        "--magneticConfig.coilConfig",
        help="set the coil config for the relative currents",
    )
    parserPoincare.add_argument(
        "--magneticConfig.path",
        default=None,
        help="create config with magnetic field grid at path",
    )
    parserPoincare.set_defaults(func=poincare)

    args = parser.parse_args()
    args.func(args)



import w7x
import numpy as np
import transcoding as tc
import tfields
import logging
from past.builtins import basestring


class mgrid_file:
    """ A set of routines to read/write mgrid files for VMEC/EXTENDER"""

    def __init__(self, parameter):
        if isinstance(parameter, basestring):
            self.read(parameter)
        else:
            self.fromWebserviceConfig(parameter)

    def toWebserviceConfig_flt(self):
        """
        Store the mgrid file contents in an equivalent field line tracer
        compatible magnetic conf datatype.
        """
        tracer = w7x.get_server(w7x.Server.addr_field_line_server)

        # Figure out the coil groups (same coilgroup and current)
        coils = []
        currents = []
        for coil in self.coils:
            # Prepare the vertices
            pf1 = tracer.types.PolygonFilament()
            pf1.vertices = tracer.types.Points3D()
            pf1.name = coil["name"]
            pf1.vertices.x1 = coil["x"].tolist()
            pf1.vertices.x2 = coil["y"].tolist()
            pf1.vertices.x3 = coil["z"].tolist()

            coils.append(pf1)
            currents.append(coil["I"][0])

        config = tracer.types.MagneticConfig()
        config.coils = coils
        config.coilsCurrents = currents

        if self.filename is None:
            config.name = "Config from mgrid_coilfile.py"
        else:
            config.name = self.filename

        # Add a grid for caching the Biot-Savart law stage.
        gridSymmetry = 5
        grid = tracer.types.Grid()
        grid.fieldSymmetry = gridSymmetry

        cyl = tracer.types.CylindricalGrid()
        cyl.numR, cyl.numZ, cyl.numPhi = 181, 181, 481
        cyl.RMin, cyl.RMax, cyl.ZMin, cyl.ZMax = 4.05, 6.75, -1.35, 1.35

        grid.cylindrical = cyl
        config.grid = grid

        return config

    def toWebserviceConfig_vmec(self):
        extender = w7x.get_server(w7x.Server.addr_extender_server)

        # Figure out the coil groups (same coilgroup and current)
        coilGroups = []
        for coil in self.coils:
            coilGroups.append([coil["coilgroup"], abs(coil["I"][0])])

        # Pick unique rows:
        coilGroups = np.array(coilGroups)
        coilGroups = np.vstack({tuple(row) for row in coilGroups})

        print("There will be {} cirquits (coil groups).".format(len(coilGroups)))
        circuits = []  # coilgroup
        coilCurrents = []
        for i in range(0, len(coilGroups)):

            sc = extender.types.SerialCircuit()
            sc.currentCarrier = []

            for coil in self.coils:
                # Pick only coils belonging to the current coil group
                if (
                    coil["coilgroup"] != int(coilGroups[i][0])
                    or np.abs(np.abs(coil["I"][0]) - np.abs(coilGroups[i][1])) > 1e-3
                ):  # noqa
                    continue

                sc.name = "coil group"
                # The current direction goes to the winding direction
                sc.current = np.abs(coil["I"][0])

                # Prepare the vertexes
                pf1 = extender.types.PolygonFilament()
                pf1.vertices = extender.types.Points3D()
                pf1.vertices.x1 = coil["x"]
                pf1.vertices.x2 = coil["y"]
                pf1.vertices.x3 = coil["z"]

                # Set up the coil parameters
                coil1 = extender.types.Coil()
                coil1.name = coil["name"]
                if coil["I"][0] > 0:
                    coil1.windingDirection = 1.0
                else:
                    coil1.windingDirection = -1.0
                coil1.numWindings = 1

                # Include the vertexes in the coil parameters
                coil1.currentCarrierPrimitive = [pf1]

                # include the coil paramaters in the coil group
                sc.currentCarrier.append(coil1)

            # include the coil group in the config
            circuits.append(sc)
            coilCurrents.append(sc.current)

        config = extender.types.MagneticConfiguration()
        config.circuit = circuits
        if self.filename is not None:
            config.name = self.filename
        else:
            config.name = "Config from mgrid_coilfile.py"

        return config, coilCurrents

    def fromWebserviceConfig(self, config):
        import numpy as np

        coils = []
        # for coilgroup in range(0,len(config))
        for coilgroup, circuit in enumerate(config.circuit):
            if np.abs(circuit.current) == 0.0:
                print(
                    "No current in circuit/coilgroup {}/{} => "
                    "skipping.".format(coilgroup, coilgroup + 1)
                )
                continue
            for coil in circuit.currentCarrier:
                name = coil.name
                nWinds = coil.numWindings
                direction = coil.windingDirection
                x = np.array(coil.currentCarrierPrimitive[0].vertices.x1).T
                y = np.array(coil.currentCarrierPrimitive[0].vertices.x2).T
                z = np.array(coil.currentCarrierPrimitive[0].vertices.x3).T
                I = circuit.current * direction * nWinds * np.ones_like(x)  # noqa
                I[-1] = 0.0
                coils.append(
                    {
                        "x": x,
                        "y": y,
                        "z": z,
                        "I": I,
                        "coilgroup": coilgroup + 1,
                        "name": name,
                    }
                )
        self.coils = coils
        self.filename = None
        self.nPeriods = 5

    def write(self, filename):
        with open(filename, "w") as f:
            # Header
            f.write("periods {}\n".format(self.nPeriods))
            f.write("begin filament\n")
            f.write("mirror NULL")  # The new line comes from next print

            # Each coil
            for c in self.coils:
                for i in range(0, len(c["x"])):
                    f.write(
                        "\n{: 12.8f} {: 12.8f} {: 12.8f} {: 15.8e}".format(
                            c["x"][i], c["y"][i], c["z"][i], c["I"][i]
                        )
                    )
                f.write(" {} {}".format(c["coilgroup"], c["name"]))
            f.write("\nend\n")

    def read(self, filename):
        import numpy as np

        with open(filename, "r") as f:

            # Read number of periods.
            str = f.readline()
            if str[0:8] != "periods ":
                raise ValueError('Missing "periods " in the beginning of file.')
            nPeriods = int(str[8:])

            # Read begin filament
            str = f.readline()
            if str[0:14] != "begin filament":
                raise ValueError('Missing "begin filament" in the beginning of file.')

            # Read/skip "mirror NULL"
            f.readline()

            # Read coils
            coils = []
            while True:
                # Read segments
                x = np.array([], float)
                y = np.array([], float)
                z = np.array([], float)
                I = np.array([], float)  # noqa
                while True:
                    strs = f.readline().split()
                    v = [float(strs[i]) for i in range(0, 4)]
                    x = np.append(x, v[0])
                    y = np.append(y, v[1])
                    z = np.append(z, v[2])
                    I = np.append(I, v[3])  # noqa

                    # Each coil is marked to end with a 0 Amps segment
                    if v[3] == 0.0:
                        coilgroup = int(strs[4])
                        name = strs[5:]
                        coils.append(
                            {
                                "x": x,
                                "y": y,
                                "z": z,
                                "I": I,
                                "coilgroup": coilgroup,
                                "name": name,
                            }
                        )
                        break
                # Check if we are at the end of file
                last_pos = f.tell()
                str = f.readline()
                if str[0:3].lower() == "end":
                    break
                f.seek(last_pos)  # Go back where we were.
        self.filename = filename
        self.coils = coils
        self.nPeriods = nPeriods

    def scaleCurrents(self, scale):
        for i in range(0, len(self.coils)):
            self.coils[i]["I"] = scale * self.coils[i]["I"]
        if not hasattr(self, "currentScalingHistory"):
            self.currentScalingHistory = []
        self.currentScalingHistory.append(scale)


class Run(object):
    def __init__(self, vmec_id):
        self._vmec_run = None
        self.vmec_run = w7x.vmec.Run(vmec_id)

    @property
    def vmec_run(self):
        return self._vmec_run

    @vmec_run.setter
    def vmec_run(self, vmec_run):
        self._vmec_run = vmec_run

    def get_coils(self):
        """
        Mostly taken from joachim geiger
        """
        log = logging.getLogger()
        extenderServer = w7x.get_server(w7x.Server.addr_extender_server)
        coilsDbServer = w7x.get_server(w7x.Server.addr_coils_db)
        coilCurrents = self.vmec_run.magnetic_config.coil_currents("A")
        # read the W7-X coils data from CoilsDB...

        # coilCurrents[0:5] = non-planar coils
        # coilCurrents[5:7] = planar coils
        # coilCurrents[7]   = sweep coils half-modules 0
        # coilCurrents[8]   = sweep coils half-modules 1

        if len(coilCurrents) > 7:
            print("Using also sweep coils")
            sweepCoils = True
        else:
            sweepCoils = False

        for i in range(7):
            coilCurrents[i] = -coilCurrents[i]

        log.info("Generate MagneticConfiguration from CoilsDB!")
        log.info("Define Coils")
        # data used in the codes lowendel and w7
        # coils collected as one type, i.e. having the same coil current
        # -------------------------
        type1 = [0, 5, 10, 15, 20, 25, 30, 35, 40, 45]
        type2 = [1, 6, 11, 16, 21, 26, 31, 36, 41, 46]
        type3 = [2, 7, 12, 17, 22, 27, 32, 37, 42, 47]
        type4 = [3, 8, 13, 18, 23, 28, 33, 38, 43, 48]
        type5 = [4, 9, 14, 19, 24, 29, 34, 39, 44, 49]
        typeA = [50, 52, 54, 56, 58, 60, 62, 64, 66, 68]
        typeB = [51, 53, 55, 57, 59, 61, 63, 65, 67, 69]
        # mirrored coils
        # -------------------------
        type1m = [80, 85, 90, 95, 100, 105, 110, 115, 120, 125]
        type2m = [81, 86, 91, 96, 101, 106, 111, 116, 121, 126]
        type3m = [82, 87, 92, 97, 102, 107, 112, 117, 122, 127]
        type4m = [83, 88, 93, 98, 103, 108, 113, 118, 123, 128]
        type5m = [84, 89, 94, 99, 104, 109, 114, 119, 124, 129]
        typeAm = [130, 132, 134, 136, 138, 140, 142, 144, 146, 148]
        typeBm = [131, 133, 135, 137, 139, 141, 143, 145, 147, 149]

        # sweep coils 0-halfmodules and 1-halfmodules
        # --------------------------
        sweep0 = range(150, 160, 2)
        sweep1 = range(151, 160, 2)

        # assure stellarator symmetry by combining coils with mirrored coils
        # -------------------------
        l_sym = True
        # combination of coils and mirrored coils to ensure stellarator symmetry
        # -------------------------
        if l_sym:
            type1 = type1 + type1m
            type2 = type2 + type2m
            type3 = type3 + type3m
            type4 = type4 + type4m
            type5 = type5 + type5m
            typeA = typeA + typeAm
            typeB = typeB + typeBm

        # generate list of the coil types used.
        # -------------------------
        coilIds = [type1, type2, type3, type4, type5, typeA, typeB]
        if sweepCoils:
            coilIds = coilIds + [sweep0, sweep1]
        circuits = []
        # i=0
        # for c in coilIds:
        #     if len(c) == 20 :
        #         coilCurrents[i]=coilCurrents[i]/2.0
        #     i=i+1

        log.info(" Retrieve Coil geometries from CoilsDB ")
        log.info(" and build up MagneticConfiguration ")
        """
        Explanation to current implementation:
           The current implementation is not quite as it consistent as it could
           be but it is the most convenient one. Each filament is treated a
           coil and not as one might like when being a purist as a filament. In
           the later sense, a coil would consist of two filaments and the
           currentCarrierRatios could be used correctly.  To do this the
           filaments of the typexm coils would have to be extracted and to be
           added either to the typex-coils or a new twofilament coil would have
           to be built up.  The current treatment uses the convenience of
           adding the two coil setups by concatenating one after the other.
           Therefore the coil current has to be divided by 2 in order to get
           the correct field strength. The currentCarrierRatios are commented
           presently (might change with a more consistent treatment in future).
        """

        for i in range(0, len(coilIds)):
            sc = extenderServer.types.SerialCircuit()
            sc.name = "coil group"
            sc.current = coilCurrents[i]
            sc.currentCarrier = []
            coils = coilsDbServer.service.getCoilData(coilIds[i])
            j = 0
            for c in coils:
                pf1 = extenderServer.types.PolygonFilament()
                pf1.vertices = extenderServer.types.Points3D()
                pf1.vertices.x1 = c.vertices.x1
                pf1.vertices.x2 = c.vertices.x2
                pf1.vertices.x3 = c.vertices.x3
                coil1 = extenderServer.types.Coil()
                coil1.name = c.name
                coil1.windingDirection = 1.0
                if i < 5:
                    coil1.numWindings = 108
                    # Mirrored coils have half the windings
                    if len(type1) == 20:
                        coil1.numWindings = 54
                        # sc.current = coilCurrents[i]
                        if j > 9:
                            coil1.windingDirection = -1.0
                elif i < 7:
                    coil1.numWindings = 36
                    # Mirrored coils have half the windings
                    if len(typeA) == 20:
                        coil1.numWindings = 18
                        # sc.current = coilCurrents[i]
                        if j > 9:
                            coil1.windingDirection = -1.0
                else:
                    # sweep coils
                    coil1.numWindings = 8

                j += 1
                coil1.currentCarrierPrimitive = [pf1]
                sc.currentCarrier.append(coil1)

            circuits.append(sc)

        log.info(" Retrieval of Coil geometries from CoilsDB done ")
        # use the coil data for a 'getExtendedField' call ...
        # -------------------------
        log.info(" ... and finalize MagneticConfiguration for extender call ")

        my_config = extenderServer.types.MagneticConfiguration()
        my_config.circuit = circuits
        if sweepCoils:
            my_config.name = "w7x_sweepCoils"
        else:
            my_config.name = "w7x"

        log.info("Finalization of MagneticConfiguration done ")

        return my_config


def writeCoils(vmec_id, coil_file="wout.coil"):
    w7x_config = Run(vmec_id).get_coils()
    G = mgrid_file(w7x_config)
    # invert currents
    print("Scaling the coil currents by -1")
    G.scaleCurrents(-1.0)
    G.write(coil_file)


def write_control_file(control_file="grid_in"):
    gridFile = (
        "nr 173   # number of radial grid points\n"
        "nz 170   # number of vertical grid points\n"
        "nphi 141  # number of toroidal cut planes, never be fooled to believe"
        " you can save too much on nphi\n"
        "rmin 3.810000 # optional param., can be computed autom. from"
        " boundary\n"
        "rmax 6.870000 # optional parameter\n"
        "zmax 1.450000 # optional parameter\n"
    )

    with open(control_file, "w") as f:
        f.write(gridFile)


def write_parameters_file(
    vmec_calc_url, vmec_id, param_file="parameters.txt", net_cdf_file="wout.nc"
):
    with open(param_file, "w") as f:
        f.write(vmec_calc_url + "\n")
        f.write(vmec_id + "\n")
        f.write(net_cdf_file + "\n")


def read_cylindrical(hybrid_path):
    hybridFileBox = tc.Block(
        "{header}",
        "cyl.RMin, cyl.RMax = {RMin:f}, {RMax:f}",
        "cyl.ZMin, cyl.ZMax = {ZMin:f}, {ZMax:f}",
        "cyl.numR, cyl.numZ, cyl.numPhi = {numR:d}, {numZ:d}, {numPhi:d}",
    )
    transc = tc.Transcoding(hybridFileBox)
    hybridPars = transc.read(hybrid_path)
    hybridPars.pop("header")
    cyl = w7x.flt.CylindricalGrid()
    cyl.__dict__.update(hybridPars)
    return cyl


def write_run_command(
    vmec_id,
    vmec_calc_url,
    net_cdf_file="wout.nc",
    script_file="runExtender.cmd",
    n_proc=32,
    control_file="grid_in",
    coil_file="w7x.coil",
):
    script = (
        "#!/bin/tcsh\n"
        "\n"
        "# Stdout and Sterr redirection\n"
        "#SBATCH -o ./{vmec_id}.o%j\n"
        "#SBATCH -e ./{vmec_id}.e%j\n"
        "# Initial working directory:\n"
        "#SBATCH -D ./\n"
        "# Job Name :\n"
        "#SBATCH -J extender-{vmec_id}\n"
        "# Queue (Partition):\n"
        "#SBATCH --partition=express\n"
        "# Number of nodes and MPI tasks per node:\n"
        "#SBATCH --nodes=20\n"
        "#SBATCH --ntasks-per-node=64\n"
        "# Enable Hyperthreading:\n"
        "#SBATCH --ntasks-per-core=2\n"
        "#\n"
        "# Request 504 GB of main Memory per node in Units of MB:\n"
        "#-- BATCH --mem=512000\n"
        "#\n"
        "#SBATCH --mail-type=all\n"
        "#SBATCH --mail-user=dboe@rzg.mpg.de\n"
        "# Wall clock limit:\n"
        "#SBATCH --time=0:30:00\n"
        "\n"
        "echo Start...\n"
        "\n"
        "# Run the program:\n"
        "\n"
        "set BIN=/u/mad/bin/EXTENDER_P\n"
        "\n"
        "set RUN={vmec_id}\n"
        "\n"
        "echo Downloading netcdf.\n"
        "wget {netCDFurl}\n"
        "\n"
        "set NETCDFfile={net_cdf_file}\n"
        "set COILfile={coil_file}\n"
        "set OUTPUT_SUFFIX=$RUN\n"
        "set CONTROL_FILE={control_file}\n"
        "\n"
        "set OTHER_OPTS=-full\n"
        "\n"
        "\n"
        "module load intel/16.0\n"
        "module load mkl/11.3\n"
        "module load impi/5.1.3\n"
        "module load ddt/7.0\n"
        "module load nag_flib/intel/mk24\n"
        "module load hdf5-serial/1.8.18\n"
        "module load netcdf-serial/4.4.1.1\n"
        "module load anaconda/2/4.4.0\n"
        "\n"
        "echo Starting extender with srun\n"
        "date\n"
        "\n"
        "echo srun $BIN -vmec_nyquist $NETCDFfile -i $CONTROL_FILE -c $COILfile"
        " -s $OUTPUT_SUFFIX $OTHER_OPTS\n"
        "srun $BIN -vmec_nyquist $NETCDFfile -i $CONTROL_FILE -c $COILfile -s"
        " $OUTPUT_SUFFIX $OTHER_OPTS > extender.out\n"
        "\n"
        "date\n"
        "\n"
        "echo Extender finished\n"
        "\n"
        "echo 'Starting conversion to ASCOT format'\n"
        "../to_ascoth5_onHGW.py > toascot.out\n"
        "\n"
        "echo 'Starting conversion to field line tracer format'\n"
        "../ascoth5_to_fieldLineTracer.py $RUN.h5 $RUN.dat > totracer.out\n"
        "\n".format(jobname=vmec_id, netCDFurl=vmec_calc_url, **locals())
    )

    with open(script_file, "w") as f:
        f.write(script)


def extend(args):
    import subprocess

    log = logging.getLogger()
    vmec_calc_url = (
        "http://svvmec1.ipp-hgw.mpg.de:8080/vmecrest/v1/run/"
        + args.vmec_id
        + "/wout.nc"
    )
    coil_file = args.vmec_id + ".coil"
    net_cdf_file = "wout.nc"
    script_file = "runExtender.cmd"

    work_dir = "~/Ascott/{args.vmec_id}".format(**locals())
    tfields.lib.in_out.mkdir(work_dir, isDir=True)
    with tfields.lib.in_out.cd_tmp(work_dir):
        log.info("Writing '{script_file}'.".format(**locals()))
        write_run_command(
            args.vmec_id,
            vmec_calc_url,
            net_cdf_file=net_cdf_file,
            script_file=script_file,
            n_proc=32,
            coil_file=coil_file,
        )

        log.info("Making the coil file '{coil_file}'.".format(**locals()))
        writeCoils(args.vmec_id, coil_file=coil_file)

        log.info("Making the grid/control file.")
        write_control_file()

        log.info("Making the parameter file")
        write_parameters_file(vmec_calc_url, args.vmec_id, net_cdf_file=net_cdf_file)

        log.info("Submitting it...")
        subprocess.call("sbatch " + script_file, shell=True)
        log.info("  and now we wait...")



from functools import wraps
import errno
import os
import signal


class TimeoutError(Exception):
    pass


def timeout(seconds=None, error_message=os.strerror(errno.ETIME)):
    """
    Decorator that raises TimeoutError if given time is exceeded.
    Args:
        seconds(int): time to wait until triggering the exception
        error_message(str)
    Examples:
        >>> import time
        >>> import w7x

        >>> def short_runtime():
        ...     print("Done")

        >>> def long_runtime():
        ...     time.sleep(2)
        ...     print("Done")

        No timeout
        >>> _ = w7x.lib.timeout.timeout(None)(short_runtime)()
        Done

        Timeout a long running function with the default expiry of 1 second.
        >>> _ = w7x.lib.timeout.timeout(1)(long_runtime)()  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        TimeoutError: Timer expired

    """
    if seconds is None:
        seconds = 0

    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        def wrapper(*args, **kwargs):
            if os.name == 'nt':  # signal.SIGALRM not existing under windows :/
                result = func(*args, **kwargs)
            else:
                signal.signal(signal.SIGALRM, _handle_timeout)
                signal.alarm(seconds)
                try:
                    result = func(*args, **kwargs)
                finally:
                    signal.alarm(0)
            return result

        return wraps(func)(wrapper)

    return decorator


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import os
import numpy as np
import tfields
from w7x import GeoSet


class Server:
    addr_field_line_server = "http://esb.ipp-hgw.mpg.de:8280/services/FieldLineProxy?wsdl"  # cluster alle esb  # NOQA
    # addr_mesh_server = "http://esb.ipp-hgw.mpg.de:8280/services/MeshSrv?wsdl"
    # addr_mesh_server = "http://lxpowerboz.ipp-hgw.mpg.de:88/services/cpp/MeshSrv?wsdl"
    # addr_mesh_server = "http://lxpowerboz:88/services/cpp/MeshSrv?wsdl"
    addr_mesh_server = "http://esb.ipp-hgw.mpg.de:8280/services/MeshSrv?wsdl"
    addr_vmec_server = "http://esb.ipp-hgw.mpg.de:8280/services/vmec_v8?wsdl"
    addr_extender_server = "http://esb.ipp-hgw.mpg.de:8280/services/Extender?wsdl"
    addr_components_db = (
        "http://esb.ipp-hgw.mpg.de:8280/services/ComponentsDBProxy?wsdl"
    )
    addr_coils_db = "http://esb.ipp-hgw.mpg.de:8280/services/CoilsDBProxy?wsdl"
    addr_w7xfp_server = "http://esb.ipp-hgw.mpg.de:8280/services/w7xfp?wsdl"


class Coils:
    # ideal coils ids
    idealNonPlanarCoilsIds = list(
        range(160, 210)
    )  # non planar coils (5 each half module)
    idealPlanarCoilsIds = list(
        range(210, 230)
    )  # planar A and B for all 10 half modules
    idealSweepCoilsIds = list(range(230, 240))  # sweep coils for all 10 half modules
    idealTrimCoilsIds = [
        350,
        241,
        351,
        352,
        353,
    ]  # trim coils in modules 1, 2(typeB), 3, 4, 5
    idealCoilsIds = (
        idealNonPlanarCoilsIds
        + idealPlanarCoilsIds
        + idealSweepCoilsIds
        + idealTrimCoilsIds
    )

    # ideal deformed coils. They fit best to M. Otte fluxsurface measurements
    # in op11 for highIota and jConfig. Half of the planed current is used here.
    deformedNonPlanarCoilsIds = list(range(1012, 1062))
    deformedPlanarCoilsIds = list(range(1062, 1082))
    deformedCoilsIds = (
        deformedNonPlanarCoilsIds
        + deformedPlanarCoilsIds
        + idealSweepCoilsIds
        + idealTrimCoilsIds
    )  # No Sweep and Trim coils provided here!

    # coils calculated by T. Andreva. Asbuilt with deformations
    asbuiltNonPlanarCoilsIds = list(range(1152, 1202))
    asbuiltPlanarCoilsIds = list(range(1202, 1222))
    asbuiltCoilsIds = (
        asbuiltNonPlanarCoilsIds
        + asbuiltPlanarCoilsIds
        + idealSweepCoilsIds
        + idealTrimCoilsIds
    )  # No Sweep and Trim coils provided here!


class MeshedModelsIds(object):
    meshEnd = GeoSet([164], "meshEnd", "black")
    divertor = GeoSet([165, 166, 167, 168, 169], "divertor", "firebrick")
    divertorKisslingerSagged = GeoSet(range(50, 60), "Sagged")
    divertorKisslinger = GeoSet(range(30, 70), "divertorKisslinger")
    # Top and bottom baffles
    baffle = GeoSet([320, 321, 322, 323, 324], "baffle", "black")
    #  Top and bottom toroidal closures
    tda = GeoSet([325, 326, 327, 328, 329], "tda", "dimgrey")
    # Heat shield - wall protection with carbon tiles
    shield = GeoSet([330, 331, 332, 333, 334], "shield", "slategrey")
    # Steel panels - wall protection with stainless steel
    panel = GeoSet([335, 336, 337, 338, 339], "panel", "grey")
    # Inner plasma vessel
    vessel = GeoSet([340, 341, 342, 343, 344], "vessel", "dimgrey")
    scraper = GeoSet(
        [349, 350, 351, 352, 353, 354, 355, 356, 357, 358], "scraper", "darkslateblue"
    )
    # limiters as used in the mapping for ir. Closest to op1.1
    limiterOP11 = GeoSet([489, 488, 487, 486, 485], "limiterOP11", "firebrick")
    # EPS2014 here the tiles are not existing. one greate plane
    limiter = GeoSet([479, 480, 481, 482, 483], "limiter", "firebrick")


class Defaults:
    fieldSymmetry = 5

    class VMEC:
        maxIterationsPerSequence = 40000
        maxToroidalMagneticFlux = -2.64726e00
        timeStep = 0.9
        numGridPointsRadial = [4, 9, 28, 99]
        forceToleranceLevels = [1e-3, 1e-5, 1e-9, 1e-14]
        pressureScale = 1.0
        totalToroidalCurrent = 0.0
        numModesPoloidal = 12
        numModesToroidal = 12
        ntheta = 36

    class CartesianGrid:
        numX = 500
        numY = 500
        numZ = 100
        ZMin = -1.5
        ZMax = 1.5
        XMin = -7
        XMax = 7
        YMin = -7
        YMax = 7

    class CylindricalGrid:
        RMin = 4.05
        RMax = 6.75
        ZMin = -1.35
        ZMax = 1.35
        numR = 181
        numZ = 181
        numPhi = 96 * 5 + 1
        PhiMin = None
        PhiMax = None

    class MagneticConfig:
        standard_rw = [1.0, 1.0, 1.0, 1.0, 1.0] + [0.0] * (2 + 2 + 5)
        high_iota_rw = [1.0, 1.0, 1.0, 1.0, 1.0, -0.23, -0.23] + [0.0] * (2 + 5)
        low_iota_rw = [1.0, 1.0, 1.0, 1.0, 1.0, 0.25, 0.25] + [0.0] * (2 + 5)
        inward_shifted_rw = [0.96, 0.95, 0.97, 1.07, 1.08, 0.1, -0.2] + [0.0] * (
            2 + 5
        )  # NOQA
        outward_shifted_rw = [1.04, 1.04, 1.01, 0.96, 0.96, -0.14, 0.14] + [0.0] * (
            2 + 5
        )  # NOQA
        low_mirror_rw = [0.94, 0.98, 0.98, 1.06, 1.06] + [0.0] * (2 + 2 + 5)
        high_mirror_rw = [1.08, 1.05, 1.0, 0.95, 0.92] + [0.0] * (2 + 2 + 5)
        low_shear_rw = [1.13, 1.12, 1.05, 0.85, 0.84, -0.2, 0.2] + [0.0] * (
            2 + 5
        )  # NOQA
        limiter_case_rw = [1.07, 1.10, 1.02, 0.92, 0.89, -0.1, 0.2] + [0.0] * (
            2 + 5
        )  # NOQA
        high_iota_a_rw = [1.0, 1.0, 1.0, 1.0, 1.0, -0.25, 0.0] + [0.0] * (2 + 5)
        high_iota_b_rw = [1.0, 1.0, 1.0, 1.0, 1.0, 0.0, -0.25] + [0.0] * (2 + 5)
        low_iota_a_rw = [1.0, 1.0, 1.0, 1.0, 1.0, 0.25, 0.0] + [0.0] * (2 + 5)
        low_iota_b_rw = [1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.25] + [0.0] * (2 + 5)

        scale = 1.43e6
        coilsIds = Coils.idealCoilsIds
        n_windings_npcs = 108
        n_windings_plcs = 36
        n_windings_scs = 8
        n_windings_tcs_a = 46  # 4 type A coils
        n_windings_tcs_b = 72  # 1 type B coil

    class Machine:
        mm_ids = ["divertor", "baffle", "tda", "shield", "panel", "vessel"]

    class Poincare:
        stepSize = 10e-3
        nPoints = 300
        seeds = tfields.Points3D(
            np.full((40, 3), 0.0), coord_sys=tfields.bases.CYLINDER
        )
        seeds[:20, 0] = np.linspace(5.95, 6.28, 20)
        seeds[20:25, 0] = np.linspace(5.43, 5.63, 5)
        seeds[25:30, 0] = np.linspace(5.43, 5.63, 5)
        seeds[30:35, 0] = np.linspace(5.40, 5.52, 5)
        seeds[35:40, 0] = np.linspace(5.40, 5.52, 5)
        seeds[:20, 2] = np.linspace(0.0, 0.0, 20)
        seeds[20:25, 2] = np.linspace(0.94, 0.94, 5)
        seeds[25:30, 2] = np.linspace(-0.94, -0.94, 5)
        seeds[30:35, 2] = np.linspace(0.85, 0.60, 5)
        seeds[35:40, 2] = np.linspace(-0.85, -0.60, 5)

        geometry_color = "grey"

    class Diffusion:
        diffusion = 1.0  # [m^2/s]
        velocity = 140000  # [m/s] (10 - 200 eV  -> 140000 m/s)
        meanFreePath = 0.1  # [m]
        stepSize = 5e-2  # [m]
        nRevolutions = 4000
        startPointShift = 0.0

    class Overload:
        maxLoadPumpingGapStirn = 2  # MW/m^2
        maxLoadPumpingGapDach = 5  # MW/m^2
        maxLoadCooled = 10  # MW/m^2
        maxLoadSagged = 1  # MW/m^2
        maxPeakLoads = {}  # Other Components Max loads in MW/m^2
        maxPeakLoads["baffle"] = 0.5
        maxPeakLoads["tda"] = 0.5
        maxPeakLoads["shield"] = 0.5  # only in higher loaded area, elsewhere 0.3
        maxPeakLoads["panel"] = 0.2
        maxPeakLoads["vessel"] = 0.0019
        radius = 0.022

    class Paths:
        divertor_upper_tiles_template_path = os.path.join(
            os.path.dirname(__file__), "data/cut_templates/mm_id-165-upper-tiles/"
        )
        baffle_upper_tiles_template_path = os.path.join(
            os.path.dirname(__file__), "data/cut_templates/mm_id-320-upper-tiles/"
        )
        shield_upper_tiles_template_path = os.path.join(
            os.path.dirname(__file__), "data/cut_templates/mm_id-330-upper-tiles/"
        )
        shield_tiles_template_path = os.path.join(
            os.path.dirname(__file__), "data/cut_templates/mm_id-330-tiles/"
        )
        """
        VMEC + EXTENDER standard case (0 current, 0 pressure):
        """
        test_dat_file = os.path.join(
            os.path.dirname(__file__),
            "../tests/resources/field_mfbe181x181x96.w7x.1000_1000_1000_1000_+0000_+0000.dboe_01.v_00_pres_00_it_12.dat",  # noqa
        )  # NOQA
        # full_path = tracer_network_share + tracer_network_share_relative_path
        tracer_network_share = "//share.ipp-hgw.mpg.de/mp/"
        tracer_network_share_mount = "/mnt/mp/"
        tracer_network_share_workgroup = "IPP-HGW"
        tracer_network_share_relative_path = "fieldline/w7x/"
        extenderDataPath = os.path.join("~/Data/", "EXTENDER")
        mgrid = "/u/amerlo/src/vmec/MAKEGRID/mgrid_w7x_nv36_hires.nc"
        vmec = "/u/amerlo/src/STELLOPT/VMEC2000/Release/xvmec2000"


#  TODO(@amerlo): Remove from here
class Batch:
    class DatasetConfig:
        size = 1000
        features = [
            "coilCurrents",
            "current_profile",
            "pressure_profile",
            "lmns",
            "gmnc",
            "bmnc",
            "bsubumnc",
            "bsubvmnc",
            "bsubsmns",
            "currumnc",
            "currvmnc",
            "bsubumnc_sur",
            "bsubvmnc_sur",
            "bsupumnc_sur",
            "bsupvmnc_sur",
            "bsupumnc",
            "bsupvmnc",
        ]

    class DistributionConfig:
        supported = ["normal", "multivariate_normal", "uniform"]

    class StemConfig:
        supported = ["list", "int", "float", "ndarray"]

    class GraphicConfig:
        supported = [
            "bmnc",
            "coilCurrents",
            "currentProfileCoefficients",
            "pressureProfileCoefficients",
        ]


class Hpc:
    class Draco:
        small = {
            "name": "hpc",
            "userid": "amerlo",
            "partition": "small",
            "tasks": 16,
            "nodes": 1,
            "mem": "2G",
        }

    class Cobra:
        tiny = {
            "name": "hpc",
            "userid": "amerlo",
            "partition": "tiny",
            "tasks": 20,
            "nodes": 1,
            "mem": "16G",
        }
